# test_sync.py
#
# Different test scenarios designed to run under management of a kernel

from collections import deque
from curio import *

import threading
import time

# ---- Synchronization primitives

class TestEvent:
    def test_event_get_wait(self, kernel):
        results = []
        async def event_setter(evt, seconds):
              results.append('sleep')
              await sleep(seconds)
              results.append('event_set')
              await evt.set()

        async def event_waiter(evt):
              results.append('wait_start')
              results.append(evt.is_set())
              await evt.wait()
              results.append('wait_done')
              results.append(evt.is_set())
              evt.clear()
              results.append(evt.is_set())

        async def main():
            evt = Event()
            await spawn(event_waiter(evt))
            await spawn(event_setter(evt, 1))

        kernel.run(main())
        assert results == [
                'wait_start',
                False,
                'sleep',
                'event_set',
                'wait_done',
                True,
                False
                ]

    def test_event_get_immediate(self, kernel):
        results = []
        async def event_setter(evt):
              results.append('event_set')
              await evt.set()

        async def event_waiter(evt, seconds):
              results.append('sleep')
              await sleep(seconds)
              results.append('wait_start')
              await evt.wait()
              results.append('wait_done')

        async def main():
            evt = Event()
            await spawn(event_waiter(evt, 1))
            await spawn(event_setter(evt))

        kernel.run(main())
        assert results == [
                'sleep',
                'event_set',
                'wait_start',
                'wait_done',
                ]


    def test_event_wait_cancel(self, kernel):
        results = []
        async def event_waiter(evt):
              results.append('event_wait')
              try:
                   await evt.wait()
              except CancelledError:
                   results.append('event_cancel')

        async def event_cancel(seconds):
              evt = Event()
              task = await spawn(event_waiter(evt))
              results.append('sleep')
              await sleep(seconds)
              results.append('cancel_start')
              await task.cancel()
              results.append('cancel_done')

        kernel.run(event_cancel(1))

        assert results == [
                'event_wait',
                'sleep',
                'cancel_start',
                'event_cancel',
                'cancel_done',
                ]

    def test_event_wait_timeout(self, kernel):
        results = []
        async def event_waiter(evt):
              results.append('event_wait')
              try:
                  await timeout_after(0.5, evt.wait())
              except TaskTimeout:
                   results.append('event_timeout')

        async def event_run(seconds):
              evt = Event()
              task = await spawn(event_waiter(evt))
              results.append('sleep')
              await sleep(seconds)
              results.append('sleep_done')

        kernel.run(event_run(1))

        assert results == [
                'event_wait',
                'sleep',
                'event_timeout',
                'sleep_done',
                ]

    def test_event_wait_notimeout(self, kernel):
        results = []
        async def event_waiter(evt):
              results.append('event_wait')
              try:
                  await timeout_after(1.0, evt.wait())
                  results.append('got event')
              except TaskTimeout:
                  results.append('event_timeout')
              
              evt.clear()
              try:
                  await evt.wait()
                  results.append('got event')
              except TaskTimeout:
                  results.append('bad timeout')

        async def event_run():
              evt = Event()
              task = await spawn(event_waiter(evt))
              results.append('sleep')
              await sleep(0.25)
              results.append('event_set')
              await evt.set()
              await sleep(1.0)
              results.append('event_set')
              await evt.set()

        kernel.run(event_run())
        assert results == [
                'event_wait',
                'sleep',
                'event_set',
                'got event',
                'event_set',
                'got event'
                ]

class TestLock:
    def test_lock_sequence(self, kernel):
        results = []
        async def worker(lck, label):
              results.append(label + ' wait')
              results.append(lck.locked())
              async with lck:
                   results.append(label + ' acquire')
                   await sleep(0.25)
              results.append(label + ' release')

        async def main():
            lck = Lock()
            await spawn(worker(lck, 'work1'))
            await spawn(worker(lck, 'work2'))
            await spawn(worker(lck, 'work3'))

        kernel.run(main())
        assert results == [
                'work1 wait',
                False,
                'work1 acquire',
                'work2 wait',
                True,
                'work3 wait',
                True,
                'work2 acquire',
                'work1 release',
                'work3 acquire',
                'work2 release',
                'work3 release',
                ]

    def test_lock_acquire_cancel(self, kernel):
        results = []
        async def worker(lck):
              results.append('lock_wait')
              try:
                  async with lck:
                       results.append('never here')
              except CancelledError:
                  results.append('lock_cancel')

        async def worker_cancel(seconds):
              lck = Lock()
              async with lck:
                  task = await spawn(worker(lck))
                  results.append('sleep')
                  await sleep(seconds)
                  results.append('cancel_start')
                  await task.cancel()
                  results.append('cancel_done')

        kernel.run(worker_cancel(1))

        assert results == [
                'lock_wait',
                'sleep',
                'cancel_start',
                'lock_cancel',
                'cancel_done',
                ]

    def test_lock_acquire_timeout(self, kernel):
        results = []
        async def worker(lck):
              results.append('lock_wait')
              try:
                  await timeout_after(0.5, lck.acquire())
                  results.append('never here')
                  await lck.release()
              except TaskTimeout:
                  results.append('lock_timeout')

        async def worker_timeout(seconds):
              lck = Lock()
              async with lck:
                  await spawn(worker(lck))
                  results.append('sleep')
                  await sleep(seconds)
                  results.append('sleep_done')

        kernel.run(worker_timeout(1))

        assert results == [
                'lock_wait',
                'sleep',
                'lock_timeout',
                'sleep_done',
                ]

class TestSemaphore:
    def test_sema_sequence(self, kernel):
        results = []
        async def worker(sema, label):
              results.append(label + ' wait')
              results.append(sema.locked())
              async with sema:
                   results.append(label + ' acquire')
                   await sleep(0.25)
              results.append(label + ' release')

        async def main():
            sema = Semaphore()
            await spawn(worker(sema, 'work1'))
            await spawn(worker(sema, 'work2'))
            await spawn(worker(sema, 'work3'))

        kernel.run(main())

        assert results == [
                'work1 wait',
                False,
                'work1 acquire',
                'work2 wait',
                True,
                'work3 wait',
                True,
                'work2 acquire',
                'work1 release',
                'work3 acquire',
                'work2 release',
                'work3 release',
                ]

    def test_sema_sequence2(self, kernel):
        results = []
        async def worker(sema, label, seconds):
              results.append(label + ' wait')
              results.append(sema.locked())
              async with sema:
                   results.append(label + ' acquire')
                   await sleep(seconds)
              results.append(label + ' release')

        async def main():
            sema = Semaphore(2)
            await spawn(worker(sema, 'work1', 0.25))
            await spawn(worker(sema, 'work2', 0.30))
            await spawn(worker(sema, 'work3', 0.35))

        kernel.run(main())
        assert results == [
                'work1 wait',            # Both work1 and work2 admitted
                False,
                'work1 acquire',
                'work2 wait',
                False,
                'work2 acquire',
                'work3 wait',
                True,
                'work3 acquire',
                'work1 release',
                'work2 release',
                'work3 release',
                ]

    def test_sema_acquire_cancel(self, kernel):
        results = []
        async def worker(lck):
              results.append('lock_wait')
              try:
                  async with lck:
                       results.append('never here')
              except CancelledError:
                  results.append('lock_cancel')

        async def worker_cancel(seconds):
              lck = Semaphore()
              async with lck:
                  task = await spawn(worker(lck))
                  results.append('sleep')
                  await sleep(seconds)
                  results.append('cancel_start')
                  await task.cancel()
                  results.append('cancel_done')

        kernel.run(worker_cancel(1))

        assert results == [
                'lock_wait',
                'sleep',
                'cancel_start',
                'lock_cancel',
                'cancel_done',
                ]

    def test_sema_acquire_timeout(self, kernel):
        results = []
        async def worker(lck):
              results.append('lock_wait')
              try:
                  await timeout_after(0.5, lck.acquire())
                  results.append('never here')
                  await lck.release()
              except TaskTimeout:
                  results.append('lock_timeout')

        async def worker_timeout(seconds):
              lck = Semaphore()
              async with lck:
                  await spawn(worker(lck))
                  results.append('sleep')
                  await sleep(seconds)
                  results.append('sleep_done')

        kernel.run(worker_timeout(1))

        assert results == [
                'lock_wait',
                'sleep',
                'lock_timeout',
                'sleep_done',
                ]

    def test_bounded(self, kernel):
        results = []
        async def task():
            sema = BoundedSemaphore(1)
            try:
                await sema.release()
                results.append('not here')
            except ValueError:
                results.append('value error')

        kernel.run(task())

        assert results == [
                'value error',
                ]

class TestCondition:
    def test_cond_sequence(self, kernel):
        results = []
        async def consumer(cond, q, label):
              while True:
                  async with cond:
                      if not q:
                          results.append(label + ' wait')
                          await cond.wait()
                      item = q.popleft()
                      if item is None:
                          break
                      results.append((label, item))
              results.append(label+' done') 

        async def producer(cond, q, count, nproducers):
             for n in range(count):
                 async with cond:
                     q.append(n)
                     results.append(('producing', n))
                     await cond.notify()
                 await sleep(0.1)

             for n in range(nproducers):
                 async with cond:
                     q.append(None)
                     results.append(('ending', n))
                     await cond.notify()
                 await sleep(0.1)
              
        async def main():
            cond = Condition()
            q = deque()
            await spawn(consumer(cond, q, 'cons1'))
            await spawn(consumer(cond, q, 'cons2'))
            await spawn(producer(cond, q, 4, 2))

        kernel.run(main())

        assert results == [
                'cons1 wait',
                'cons2 wait',
                ('producing', 0),
                ('cons1', 0),
                'cons1 wait',
                ('producing', 1),
                ('cons2', 1),
                'cons2 wait',
                ('producing', 2),
                ('cons1', 2),
                'cons1 wait',
                ('producing', 3),
                ('cons2', 3),
                'cons2 wait',
                ('ending', 0),
                ('cons1 done'),
                ('ending', 1),
                ('cons2 done')
                ]

    def test_cond_wait_cancel(self, kernel):
        results = []
        async def worker(cond):
              try:
                  async with cond:
                       results.append('cond_wait')
                       await cond.wait()
                       results.append('never here')
              except CancelledError:
                  results.append('worker_cancel')

        async def worker_cancel(seconds):
              cond = Condition()
              task = await spawn(worker(cond))
              results.append('sleep')
              await sleep(seconds)
              results.append('cancel_start')
              await task.cancel()
              results.append('cancel_done')

        kernel.run(worker_cancel(1))

        assert results == [
                'cond_wait',
                'sleep',
                'cancel_start',
                'worker_cancel',
                'cancel_done',
                ]

    def test_cond_wait_timeout(self, kernel):
        results = []
        async def worker(cond):
              try:
                  async with cond:
                       results.append('cond_wait')
                       await timeout_after(0.25, cond.wait())
                       results.append('never here')
              except TaskTimeout:
                  results.append('worker_timeout')

        async def worker_cancel(seconds):
              cond = Condition()
              task = await spawn(worker(cond))
              results.append('sleep')
              await sleep(seconds)
              results.append('done')

        kernel.run(worker_cancel(1))

        assert results == [
                'cond_wait',
                'sleep',
                'worker_timeout',
                'done'
                ]

    def test_cond_notify_all(self, kernel):
        results = []
        async def worker(cond):
             async with cond:
                 results.append('cond_wait')
                 await cond.wait()
                 results.append('wait_done')

        async def worker_notify(seconds):
              cond = Condition()
              await spawn(worker(cond))
              await spawn(worker(cond))
              await spawn(worker(cond))
              results.append('sleep')
              await sleep(seconds)
              async with cond:
                  results.append('notify')
                  await cond.notify_all()
              results.append('done')

        kernel.run(worker_notify(1))

        assert results == [
                'cond_wait',
                'cond_wait',
                'cond_wait',
                'sleep',
                'notify',
                'wait_done',
                'done',
                'wait_done',
                'wait_done',
                ]

    def test_cond_waitfor(self, kernel):
        results = []
        async def consumer(cond, q, label):
             async with cond:
                 results.append(label + ' waitfor')
                 await cond.wait_for(lambda: len(q) > 2)
                 results.append((label, len(q)))
             results.append(label+' done') 

        async def producer(cond, q, count):
             for n in range(count):
                 async with cond:
                     q.append(n)
                     results.append(('producing', n))
                     await cond.notify()
                 await sleep(0.1)

        async def main():              
            cond = Condition()
            q = deque()
            await spawn(consumer(cond, q, 'cons1'))
            await spawn(consumer(cond, q, 'cons2'))
            await spawn(producer(cond, q, 4))

        kernel.run(main())
        assert results == [
                'cons1 waitfor',
                'cons2 waitfor',
                ('producing', 0),
                ('producing', 1),
                ('producing', 2),
                ('cons1', 3),
                'cons1 done',
                ('producing', 3),
                ('cons2', 4),
                'cons2 done'
                ]


class TestAbide:
    def test_abide_async(self, kernel):
        results = []
        async def waiter(lck, evt):
              await abide(lck.acquire)
              results.append('acquired')
              await abide(lck.release)
              results.append('released')
              await abide(evt.set)

        async def tester(lck, evt):
              async with lck:
                  results.append('tester')
                  await sleep(0.1)
              await evt.wait()
              async with lck:
                  results.append('tester finish')

        async def main():
            lck = Lock()
            evt = Event()
            await spawn(tester(lck, evt))
            await spawn(waiter(lck, evt))

        kernel.run(main())
        assert results == [
                'tester',
                'acquired',
                'released',
                'tester finish'
                ]

    def test_abide_async_with(self, kernel):
        results = []
        async def waiter(lck, evt):
              async with abide(lck):
                  results.append('acquired')
              results.append('released')
              await abide(evt.set)

        async def tester(lck, evt):
              async with lck:
                  results.append('tester')
                  await sleep(0.1)
              await evt.wait()
              async with lck:
                  results.append('tester finish')

        async def main():
            lck = Lock()
            evt = Event()
            await spawn(tester(lck, evt))
            await spawn(waiter(lck, evt))

        kernel.run(main())
        assert results == [
                'tester',
                'acquired',
                'released',
                'tester finish'
                ]

    def test_abide_sync(self, kernel):
        results = []
        async def waiter(lck, evt):
              await abide(lck.acquire)
              results.append('acquired')
              await abide(lck.release)
              results.append('released')
              await abide(evt.set)

        # Synchronous code. Runs in a thread
        def tester(lck, evt):
              with lck:
                  results.append('tester')
                  time.sleep(0.1)
              evt.wait()
              with lck:
                  results.append('tester finish')

        async def main():
            lck = threading.Lock()
            evt = threading.Event()
            await spawn(run_in_thread(tester, lck, evt))
            await sleep(0.01)
            await spawn(waiter(lck, evt))

        kernel.run(main())
        assert results == [
                'tester',
                'acquired',
                'released',
                'tester finish'
                ]

    def test_abide_sync_with(self, kernel):
        results = []
        async def waiter(lck, evt):
              async with abide(lck):
                  results.append('acquired')
              results.append('released')
              await abide(evt.set)

        # Synchronous code. Runs in a thread
        def tester(lck, evt):
              with lck:
                  results.append('tester')
                  time.sleep(0.1)
              evt.wait()
              with lck:
                  results.append('tester finish')

        async def main():
            lck = threading.Lock()
            evt = threading.Event()
            await spawn(run_in_thread(tester, lck, evt))
            await sleep(0.01)
            await spawn(waiter(lck, evt))

        kernel.run(main())
        assert results == [
                'tester',
                'acquired',
                'released',
                'tester finish'
                ]


    def test_abide_sync_with_cancel(self, kernel):
        results = []
        async def waiter(lck, evt):
              try:
                  async with timeout_after(0.5):       
                      async with abide(lck):
                          results.append('acquired')
                      results.append('released')
              except TaskTimeout:
                  results.append('timeout')
              await abide(evt.set)

        # Synchronous code. Runs in a thread
        def tester(lck, evt):
              with lck:
                  results.append('tester')
                  time.sleep(1)
              time.sleep(0.1)
              with lck:
                  results.append('tester2')
              evt.wait()
              with lck:
                  results.append('tester finish')

        async def main():
            lck = threading.Lock()
            evt = threading.Event()
            await spawn(run_in_thread(tester, lck, evt))
            await sleep(0.01)
            await spawn(waiter(lck, evt))

        kernel.run(main())
        assert results == [
                'tester',
                'timeout',
                'tester2',
                'tester finish'
                ]



# test_queue.py

from collections import deque
from curio import *

def test_queue_simple(kernel):
    results = []
    async def consumer(queue, label):
          while True:
              item = await queue.get()
              if item is None:
                  break
              results.append((label, item))
              await queue.task_done()
          await queue.task_done()
          results.append(label + ' done')

    async def producer():
        queue = Queue()
        results.append('producer_start')
        await spawn(consumer(queue, 'cons1'))
        await spawn(consumer(queue, 'cons2'))
        await sleep(0.1)
        for n in range(4):
            await queue.put(n)
            await sleep(0.1)
        for n in range(2):
            await queue.put(None)
        results.append('producer_join')
        await queue.join()
        results.append('producer_done')

    kernel.run(producer())

    assert results == [
            'producer_start',
            ('cons1', 0),
            ('cons2', 1),
            ('cons1', 2),
            ('cons2', 3),
            'cons1 done',
            'cons2 done',
            'producer_join',
            'producer_done',
            ]

def test_queue_unbounded(kernel):
    results = []
    async def consumer(queue, label):
          while True:
              item = await queue.get()
              if item is None:
                  break
              results.append((label, item))
              await queue.task_done()
          await queue.task_done()
          results.append(label + ' done')

    async def producer():
        queue = Queue()
        results.append('producer_start')
        await spawn(consumer(queue, 'cons1'))
        await sleep(0.1)
        for n in range(4):
            await queue.put(n)
        await queue.put(None)
        results.append('producer_join')
        await queue.join()
        results.append('producer_done')

    kernel.run(producer())

    assert results == [
            'producer_start',
            ('cons1', 0),
            ('cons1', 1),
            ('cons1', 2),
            ('cons1', 3),
            'cons1 done',
            'producer_join',
            'producer_done',
            ]


def test_queue_bounded(kernel):
    results = []
    async def consumer(queue, label):
          while True:
              item = await queue.get()
              if item is None:
                  break
              results.append((label, item))
              await sleep(0.1)
              await queue.task_done()
          await queue.task_done()
          results.append(label + ' done')

    async def producer():
        queue = Queue(maxsize=2)
        results.append('producer_start')
        await spawn(consumer(queue, 'cons1'))
        await sleep(0.1)
        for n in range(4):
            await queue.put(n)
            results.append(('produced', n))
        await queue.put(None)
        results.append('producer_join')
        await queue.join()
        results.append('producer_done')

    kernel.run(producer())

    assert results == [
            'producer_start',
            ('cons1', 0),
            ('produced', 0),
            ('produced', 1),
            ('produced', 2),
            ('produced', 3),
            ('cons1', 1),
            'producer_join',
            ('cons1', 2),
            ('cons1', 3),
            'producer_done',
            'cons1 done',
            ]

def test_queue_get_cancel(kernel):
    # Make sure a blocking get can be cancelled
    results = []
    async def consumer():
          queue = Queue()
          try:
              results.append('consumer waiting')
              item = await queue.get()
              results.append('not here')
          except CancelledError:
              results.append('consumer cancelled')

    async def driver():
        task = await spawn(consumer())
        await sleep(0.5)
        await task.cancel()

    kernel.run(driver())
    assert results == [
            'consumer waiting',
            'consumer cancelled'
            ]

def test_queue_put_cancel(kernel):
    # Make sure a blocking put() can be cancelled
    results = []

    async def producer():
        queue = Queue(1)
        results.append('producer_start')
        await queue.put(0)
        try:
            await queue.put(1)
            results.append('not here')
        except CancelledError:
            results.append('producer_cancel')

    async def driver():
        task = await spawn(producer())
        await sleep(0.5)
        await task.cancel()

    kernel.run(driver())
    assert results == [
            'producer_start',
            'producer_cancel'
            ]

def test_queue_get_timeout(kernel):
    # Make sure a blocking get respects timeouts
    results = []
    async def consumer():
          queue = Queue()
          try:
              results.append('consumer waiting')
              item = await timeout_after(0.5, queue.get())
              results.append('not here')
          except TaskTimeout:
              results.append('consumer timeout')

    kernel.run(consumer())
    assert results == [
            'consumer waiting',
            'consumer timeout'
            ]

def test_queue_put_timeout(kernel):
    # Make sure a blocking put() respects timeouts
    results = []

    async def producer():
        queue = Queue(1)
        results.append('producer start')
        await queue.put(0)
        try:
            await timeout_after(0.5, queue.put(1))
            results.append('not here')
        except TaskTimeout:
            results.append('producer timeout')

    kernel.run(producer())
    assert results == [
            'producer start',
            'producer timeout'
            ]




# ssl_echo
#
# An example of a simple SSL echo server.   Use ssl_echo_client.py to test.

import curio
from curio import ssl
from curio import network

KEYFILE = "ssl_test_rsa"    # Private key
CERTFILE = "ssl_test.crt"   # Certificate (self-signed)

async def handle(client, addr):
    print('Connection from', addr)
    async with client:
        while True:
            data = await client.recv(1000)
            if not data:
                break
            await client.send(data)
    print('Connection closed')

if __name__ == '__main__':
    ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ssl_context.load_cert_chain(certfile=CERTFILE, keyfile=KEYFILE)
    curio.run(network.tcp_server('', 10000, handle, ssl=ssl_context))




# curio/socket.py
#
# Copyright (C) 2015
# David Beazley (Dabeaz LLC), http://www.dabeaz.com
# All rights reserved.
#
# Standin for the standard socket library.  The entire contents of stdlib socket are
# made available here.  However, the socket class is replaced by an async compatible version.
# Certain blocking operations are also replaced by versions safe to use in async.
#

import socket as _socket

__all__ = _socket.__all__

from socket import *
from functools import wraps, partial

from . import workers
from . import io

@wraps(_socket.socket)
def socket(*args, **kwargs):
    return io.Socket(_socket.socket(*args, **kwargs))

@wraps(_socket.socketpair)
def socketpair(*args, **kwargs):
    s1, s2 = _socket.socketpair(*args, **kwargs)
    return io.Socket(s1), io.Socket(s2)

@wraps(_socket.fromfd)
def fromfd(*args, **kwargs):
    return io.Socket(_socket.fromfd(*args, **kwargs))

# Replacements for blocking functions related to domain names and DNS
@wraps(_socket.create_connection)
async def create_connection(*args, **kwargs):
    sock = await workers.run_in_thread(_socket.create_connection, *args, **kwargs)
    return io.Socket(sock)

@wraps(_socket.getaddrinfo)
async def getaddrinfo(*args, **kwargs):
    return await workers.run_in_thread(_socket.getaddrinfo, *args, **kwargs)

@wraps(_socket.getfqdn)
async def getfqdn(*args, **kwargs):
    return await workers.run_in_thread(_socket.getfqdn, *args, **kwargs)

@wraps(_socket.gethostbyname)
async def gethostbyname(*args, **kwargs):
    return await workers.run_in_thread(_socket.gethostbyname, *args, **kwargs)

@wraps(_socket.gethostbyname_ex)
async def gethostbyname_ex(*args, **kwargs):
    return await workers.run_in_thread(_socket.gethostbyname_ex, *args, **kwargs)

@wraps(_socket.gethostname)
async def gethostname(*args, **kwargs):
    return await workers.run_in_thread(_socket.gethostname, *args, **kwargs)

@wraps(_socket.gethostbyaddr)
async def gethostbyaddr(*args, **kwargs):
    return await workers.run_in_thread(_socket.gethostbyaddr, *args, **kwargs)

@wraps(_socket.getnameinfo)
async def getnameinfo(*args, **kwargs):
    return await workers.run_in_thread(_socket.getnameinfo, *args, **kwargs)




    

     
    



# test_workers.py

import pytest

import time
from curio import *

def fib(n):
    if n <= 2:
        return 1
    else:
        return fib(n-1) + fib(n-2)

def test_cpu(kernel):
    results = []

    async def spin(n):
        while n > 0:
            results.append(n)
            await sleep(0.1)
            n -= 1

    async def cpu_bound(n):
         r = await run_in_process(fib, n)
         results.append(('fib', r))

    async def main():
         await spawn(spin(10))
         await spawn(cpu_bound(36))

    kernel.run(main())

    assert results == [
            10, 9, 8, 7, 6, 5, 4, 3, 2, 1,
            ('fib', 14930352)
            ]

def test_blocking(kernel):
    results = []

    async def spin(n):
        while n > 0:
            results.append(n)
            await sleep(0.1)
            n -= 1

    async def blocking(n):
         await run_in_thread(time.sleep, n)
         results.append('sleep done')

    async def main():
         await spawn(spin(10))
         await spawn(blocking(2))

    kernel.run(main())

    assert results == [
            10, 9, 8, 7, 6, 5, 4, 3, 2, 1,
            'sleep done',
            ]

@pytest.mark.parametrize('runner', [ run_in_thread, run_in_process ])
def test_worker_cancel(kernel, runner):
    results = []

    async def spin(n):
        while n > 0:
            results.append(n)
            await sleep(0.1)
            n -= 1

    async def blocking(n):
         task = await spawn(runner(time.sleep, n))
         await sleep(0.55)
         await task.cancel()
         try:
             await task.join()
         except TaskError as e:
             if isinstance(e.__cause__, CancelledError):
                 results.append('cancel')

    async def main():
         await spawn(spin(10))
         await spawn(blocking(5))

    kernel.run(main())

    assert results == [
            10, 9, 8, 7, 6, 5, 'cancel', 4, 3, 2, 1
            ]


@pytest.mark.parametrize('runner', [ run_in_thread, run_in_process ])
def test_worker_timeout(kernel, runner):
    results = []

    async def spin(n):
        while n > 0:
            results.append(n)
            await sleep(0.1)
            n -= 1

    async def blocking(n):
         try:
             result = await timeout_after(0.55, runner(time.sleep, n))
         except TaskTimeout:
             results.append('cancel')

    async def main():
         await spawn(spin(10))
         await spawn(blocking(5))

    kernel.run(main())

    assert results == [
            10, 9, 8, 7, 6, 5, 'cancel', 4, 3, 2, 1
            ]




# Example: A simple echo server written using asyncio streams

import asyncio

async def echo_client(reader, writer):
    print('Here!')
    addr = writer.get_extra_info('peername')
    print('Connection from', addr)
    while True:
         data = await reader.read(10000)
         if not data:
             break
         writer.write(data)
         await writer.drain()
    print('Connection closed')

if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    coro = asyncio.start_server(echo_client, '127.0.0.1', 25000, loop=loop)    
    loop.run_until_complete(coro)
    loop.run_forever()



# subproc_perf.py

from curio import *
from curio.subprocess import check_output
import time
import subprocess
import asyncio

COUNT = 1000

input = (b'aaa '*10 + b'\n')*10000
cmd = ['cat']

async def main(n):
    for x in range(n):
        out = await check_output(cmd, input=input)
    assert out == input

def curio_test(n):
    start = time.time()
    run(main(n))
    end = time.time()
    print('curio:', end-start)

def subprocess_test(n):
    start = time.time()
    for x in range(n):
        out = subprocess.check_output(cmd, input=input)
    assert out == input
    end = time.time()
    print('subprocess:', end-start)

def asyncio_test(n):
    async def main(n):
        for x in range(n):
            proc = await asyncio.create_subprocess_exec(*cmd, 
                                                        stdin=asyncio.subprocess.PIPE, 
                                                        stdout=asyncio.subprocess.PIPE)
            stdout, stderr = await proc.communicate(input=input)
            await proc.wait()
        assert stdout == input

    loop = asyncio.get_event_loop()
    start = time.time()
    loop.run_until_complete(asyncio.ensure_future(main(n)))
    end = time.time()
    print('asyncio:', end-start)


def uvloop_test(n):
    try:
        import uvloop
    except ImportError:
        return

    async def main(n):
        for x in range(n):
            proc = await asyncio.create_subprocess_exec(*cmd, 
                                                        stdin=asyncio.subprocess.PIPE, 
                                                        stdout=asyncio.subprocess.PIPE)
            stdout, stderr = await proc.communicate(input=input)
            await proc.wait()
        assert stdout == input

    loop = uvloop.new_event_loop()
    asyncio.set_event_loop(loop)
    start = time.time()
    loop.run_until_complete(asyncio.ensure_future(main(n)))
    end = time.time()
    print('uvloop:', end-start)
            
if __name__ == '__main__':
    curio_test(COUNT)
    subprocess_test(COUNT)
    asyncio_test(COUNT)
    uvloop_test(COUNT)







# Example: A simple Unix  echo server

from curio import run, unix_server

async def echo_handler(client, address):
    print('Connection from', address)
    while True:
         data = await client.recv(10000)
         if not data:
              break
         await client.sendall(data)
    print('Connection closed')

if __name__ == '__main__':
     import os
     try:
          os.remove('/tmp/curiounixecho')
     except:
          pass
     run(unix_server('/tmp/curiounixecho', echo_handler))
          



import asyncio
from socket import *

class EchoProtocol(asyncio.Protocol):
    def connection_made(self, transport):
        self.transport = transport
        sock = transport.get_extra_info('socket')
        try:
            sock.setsockopt(IPPROTO_TCP, TCP_NODELAY, 1)
        except (OSError, NameError):
            pass

    def connection_lost(self, exc):
        self.transport = None

    def data_received(self, data):
        self.transport.write(data)

if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    coro = loop.create_server(EchoProtocol, '', 25000)
    srv = loop.run_until_complete(coro)
    loop.run_forever()



# A hello world program. From the Curio tutorial at
# https://curio.readthedocs.org/en/latest/tutorial.html
#
import curio
import signal
import os

async def countdown(n):
    while n > 0:
        print('T-minus', n)
        await curio.sleep(1)
        n -= 1

start_evt = curio.Event()

def fib(n):
    if n <= 2:
        return 1
    else:
        return fib(n-1) + fib(n-2)

async def kid():
    while True:
        try:
            print('Can I play?')
            await curio.timeout_after(1, start_evt.wait())
            break
        except curio.TaskTimeout:
            print('Wha!?!')
    try:
        print('Building the Millenium Falcon in Minecraft')
        total = 0
        for n in range(50):
             total += await curio.run_in_process(fib, n)
    except curio.CancelledError:
        print('Fine. Saving my work. I got to', total)

async def parent():
    print('Parent PID', os.getpid())
    kid_task = await curio.spawn(kid())
    await curio.sleep(5)
    print("Yes, go play")
    await start_evt.set()

    print("Parent says send me a SIGHUP when you want me to leave with the kid")
    await curio.SignalSet(signal.SIGHUP).wait()

    print("Let's go")
    count_task = await curio.spawn(countdown(10))
    await count_task.join()
    print("We're leaving!")
    try:
        await curio.timeout_after(10, kid_task.join())
    except curio.TaskTimeout:
        print('I warned you!')
        await kid_task.cancel()
    print("Leaving!")

if __name__ == '__main__':
    curio.run(parent(), with_monitor=True)



# traps.py
#
# Coroutines corresponding to the kernel traps.  These functions
# provide the bridge from tasks to the underlying kernel. This is the
# only place where the explicit @coroutine decorator needs to be used.
# All other code in curio and in user-tasks should use async/await
# instead.  Direct use by users is allowed, but if you're working with
# these traps directly, there is probably a higher level interface
# that simplifies the problem you're trying to solve (e.g., Socket,
# File, objects, etc.)
# ----------------------------------------------------------------------

__all__ = [
    '_read_wait', '_write_wait', '_future_wait', '_sleep',
    '_spawn', '_cancel_task', '_join_task', '_wait_on_queue',
    '_reschedule_tasks', '_sigwatch', '_sigunwatch', '_sigwait',
    '_get_kernel', '_get_current', '_set_timeout', '_unset_timeout'
    ]

from types import coroutine
from selectors import EVENT_READ, EVENT_WRITE

from .errors import _CancelRetry, CancelledError

@coroutine
def _read_wait(fileobj):
    '''
    Wait until reading can be performed.
    '''
    yield ('_trap_io', fileobj, EVENT_READ, 'READ_WAIT')

@coroutine
def _write_wait(fileobj):
    '''
    Wait until writing can be performed.
    '''
    yield ('_trap_io', fileobj, EVENT_WRITE, 'WRITE_WAIT')
    
@coroutine
def _future_wait(future, event=None):
    '''
    Wait for the result of a Future to be ready.
    '''
    yield ('_trap_future_wait', future, event)

@coroutine
def _sleep(seconds):
    '''
    Sleep for a given number of seconds. Sleeping for 0 seconds
    forces the current task to yield to the next task (if any).
    '''
    yield ('_trap_sleep', seconds)

@coroutine
def _spawn(coro, daemon):
    '''
    Create a new task. Returns the resulting Task object.
    '''
    return (yield '_trap_spawn', coro, daemon)

@coroutine
def _cancel_task(task):
    '''
    Cancel a task. Causes a CancelledError exception to raise in the task.
    The exception can be changed by specifying exc. 
    '''
    while True:
        try:
            yield ('_trap_cancel_task', task)
            return
        except _CancelRetry:
            pass

@coroutine
def _join_task(task):
    '''
    Wait for a task to terminate.
    '''
    yield ('_trap_join_task', task)

@coroutine
def _wait_on_queue(queue, state):
    '''
    Put the task to sleep on a queue.
    '''
    yield ('_trap_wait_queue', queue, state)

@coroutine
def _reschedule_tasks(queue, n=1, value=None, exc=None):
    '''
    Reschedule one or more tasks waiting on a kernel queue.
    '''
    yield ('_trap_reschedule_tasks', queue, n, value, exc)

@coroutine
def _sigwatch(sigset):
    '''
    Start monitoring a signal set
    '''
    yield ('_trap_sigwatch', sigset)

@coroutine
def _sigunwatch(sigset):
    '''
    Stop watching a signal set
    '''
    yield ('_trap_sigunwatch', sigset)

@coroutine
def _sigwait(sigset):
    '''
    Wait for a signal to arrive.
    '''
    yield ('_trap_sigwait', sigset)

@coroutine
def _get_kernel():
    '''
    Get the kernel executing the task.
    '''
    result = yield ('_trap_get_kernel',)
    return result

@coroutine
def _get_current():
    '''
    Get the currently executing task
    '''
    result = yield ('_trap_get_current',)
    return result

@coroutine
def _set_timeout(seconds):
    '''
    Set a timeout for the current task.  
    '''
    result = yield ('_trap_set_timeout', seconds)
    return result

@coroutine
def _unset_timeout(previous):
    '''
    Restore the previous timeout for the current task.
    '''
    yield ('_trap_unset_timeout', previous)



# test_socket.py

from curio import *
from curio.socket import *

def test_tcp_echo(kernel):
    results = []
    async def server(address):
        sock = socket(AF_INET, SOCK_STREAM)
        sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
        sock.bind(address)
        sock.listen(5)
        results.append('accept wait')
        client, addr = await sock.accept()
        results.append('accept done')
        await spawn(handler(client))
        await sock.close()

    async def handler(client):
        results.append('handler start')
        while True:
            results.append('recv wait')
            data = await client.recv(100)
            if not data:
                break
            results.append(('handler', data))
            await client.sendall(data)
        results.append('handler done')
        await client.close()

    async def client(address):
        results.append('client start')
        sock = socket(AF_INET, SOCK_STREAM)
        await sock.connect(address)
        await sock.send(b'Msg1')
        await sleep(0.1)
        resp = await sock.recv(100)
        results.append(('client', resp))
        await sock.send(b'Msg2')
        await sleep(0.1)
        resp = await sock.recv(100)
        results.append(('client', resp))
        results.append('client close')
        await sock.close()

    async def main():
        await spawn(server(('',25000)))
        await spawn(client(('localhost',25000)))

    kernel.run(main())

    assert results == [
            'accept wait',
            'client start',
            'accept done',
            'handler start',
            'recv wait',
            ('handler', b'Msg1'),
            'recv wait',
            ('client', b'Msg1'),
            ('handler', b'Msg2'),
            'recv wait',
            ('client', b'Msg2'),
            'client close',
            'handler done'
            ]

def test_tcp_file_echo(kernel):
    results = []
    async def server(address):
        sock = socket(AF_INET, SOCK_STREAM)
        sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
        sock.bind(address)
        sock.listen(5)
        results.append('accept wait')
        client, addr = await sock.accept()
        results.append('accept done')
        await spawn(handler(client))
        await sock.close()

    async def handler(client):
        results.append('handler start')
        stream = client.as_stream()
        async for line in stream:
            results.append(('handler', line))
            await stream.write(line)
        results.append('handler done')
        await client.close()

    async def client(address):
        results.append('client start')
        sock = socket(AF_INET, SOCK_STREAM)
        await sock.connect(address)
        stream = sock.as_stream()
        await stream.write(b'Msg1\n')
        await sleep(0.1)
        resp = await stream.read(100)
        results.append(('client', resp))
        await stream.write(b'Msg2\n')
        await sleep(0.1)
        resp = await stream.read(100)
        results.append(('client', resp))
        results.append('client close')
        await sock.close()

    async def main():
        await spawn(server(('',25000)))
        await spawn(client(('localhost',25000)))

    kernel.run(main())

    assert results == [
            'accept wait',
            'client start',
            'accept done',
            'handler start',
            ('handler', b'Msg1\n'),
            ('client', b'Msg1\n'),
            ('handler', b'Msg2\n'),
            ('client', b'Msg2\n'),
            'client close',
            'handler done'
            ]

def test_udp_echo(kernel):
    results = []
    async def server(address):
        sock = socket(AF_INET, SOCK_DGRAM)
        sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
        sock.bind(address)
        results.append('recvfrom wait')
        data, addr = await sock.recvfrom(8192)
        results.append(('server', data))
        await sock.sendto(data, addr)
        await sock.close()
        results.append('server close')

    async def client(address):
        results.append('client start')
        sock = socket(AF_INET, SOCK_DGRAM)
        results.append('client send')
        await sock.sendto(b'Msg1', address)
        data, addr = await sock.recvfrom(8192)
        results.append(('client', data))
        await sock.close()
        results.append('client close')


    async def main():
        await spawn(server(('',25000)))
        await spawn(client(('localhost',25000)))

    kernel.run(main())

    assert results == [
            'recvfrom wait',
            'client start',
            'client send',
            ('server', b'Msg1'),
            'server close',
            ('client', b'Msg1'),
            'client close'
            ]

def test_accept_timeout(kernel):
    results = []
    async def server(address):
        sock = socket(AF_INET, SOCK_STREAM)
        sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
        sock.bind(address)
        sock.listen(1)
        results.append('accept wait')
        try:
            client, addr = await timeout_after(0.5, sock.accept())
            results.append('not here')
        except TaskTimeout:
            results.append('accept timeout')
        await sock.close()

    kernel.run(server(('',25000)))

    assert results == [
            'accept wait',
            'accept timeout'
            ]

def test_accept_cancel(kernel):
    results = []
    async def server(address):
        sock = socket(AF_INET, SOCK_STREAM)
        sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
        sock.bind(address)
        sock.listen(1)
        results.append('accept wait')
        try:
            client, addr = await sock.accept()
            results.append('not here')
        except CancelledError:
            results.append('accept cancel')
        await sock.close()

    async def canceller():
         task = await spawn(server(('',25000)))
         await sleep(0.5)
         await task.cancel()

    kernel.run(canceller())
    assert results == [
            'accept wait',
            'accept cancel'
            ]

def test_recv_timeout(kernel):
    results = []
    async def server(address):
        sock = socket(AF_INET, SOCK_STREAM)
        sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
        sock.bind(address)
        sock.listen(1)
        results.append('accept wait')
        client, addr = await sock.accept()
        results.append('recv wait')
        try:
            data = await timeout_after(0.5, client.recv(8192))
            results.append('not here')
        except TaskTimeout:
            results.append('recv timeout')
        await client.close()
        await sock.close()

    async def canceller():
         task = await spawn(server(('',25000)))
         sock = socket(AF_INET, SOCK_STREAM)
         results.append('client connect')
         await sock.connect(('localhost', 25000))
         await sleep(1.0)
         await sock.close()
         results.append('client done')

    kernel.run(canceller())

    assert results == [
            'accept wait',
            'client connect',
            'recv wait',
            'recv timeout',
            'client done'
            ]

def test_recv_cancel(kernel):
    results = []
    async def server(address):
        sock = socket(AF_INET, SOCK_STREAM)
        sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
        sock.bind(address)
        sock.listen(1)
        results.append('accept wait')
        client, addr = await sock.accept()
        results.append('recv wait')
        try:
            data = await client.recv(8192)
            results.append('not here')
        except CancelledError:
            results.append('recv cancel')
        await client.close()
        await sock.close()

    async def canceller():
         task = await spawn(server(('',25000)))
         sock = socket(AF_INET, SOCK_STREAM)
         results.append('client connect')
         await sock.connect(('localhost', 25000))
         await sleep(1.0)
         await task.cancel()
         await sock.close()
         results.append('client done')

    kernel.run(canceller())

    assert results == [
            'accept wait',
            'client connect',
            'recv wait',
            'recv cancel',
            'client done'
            ]

def test_recvfrom_timeout(kernel):
    results = []
    async def server(address):
        sock = socket(AF_INET, SOCK_DGRAM)
        sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
        sock.bind(address)
        results.append('recvfrom wait')
        try:
            await timeout_after(0.5, sock.recvfrom(8192))
            results.append('not here')
        except TaskTimeout:
            results.append('recvfrom timeout')
        await sock.close()

    async def canceller():
         await spawn(server(('',25000)))
         await sleep(1.0)
         results.append('client done')

    kernel.run(canceller())

    assert results == [
            'recvfrom wait',
            'recvfrom timeout',
            'client done'
            ]

def test_recvfrom_cancel(kernel):
    results = []
    async def server(address):
        sock = socket(AF_INET, SOCK_DGRAM)
        sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
        sock.bind(address)
        results.append('recvfrom wait')
        try:
            await sock.recvfrom(8192)
            results.append('not here')
        except CancelledError:
            results.append('recvfrom cancel')
        await sock.close()

    async def canceller():
         task = await spawn(server(('',25000)))
         await sleep(1.0)
         await task.cancel()
         results.append('client done')

    kernel.run(canceller())
    
    assert results == [
            'recvfrom wait',
            'recvfrom cancel',
            'client done'
            ]

def test_buffer_into(kernel):
    from array import array
    results = []
    async def sender(s1):
        a = array('i', range(1000000))
        await s1.sendall(a)

    async def receiver(s2):
        a = array('i', (0 for n in range(1000000)))
        view = memoryview(a).cast('b')
        total = 0
        while view:
            nrecv = await s2.recv_into(view)
            if not nrecv:
                break
            total += nrecv
            view = view[nrecv:]

        results.append(a)

    s1, s2 = socketpair() 

    async def main():
        await spawn(sender(s1))
        await spawn(receiver(s2))

    kernel.run(main())
    s1._socket.close()
    s2._socket.close()

    assert all(n==x for n,x in enumerate(results[0]))



# curio/monitor.py
#
# Copyright (C) 2015-2016
# David Beazley (Dabeaz LLC), http://www.dabeaz.com
# All rights reserved.
#
# Debugging monitor for curio. To enable the monitor, create a kernel
# with the with_monitor argument:
#
#    k = Kernel(with_monitor=True)
#
# Or run a curio program with the CURIOMONITOR environment variable set
#
#    env CURIOMONITOR=TRUE python3 someprog.py
#
# If you need to change some aspect of the monitor configuration, you
# can do manual setup:
#
#    k = Kernel()
#    Monitor(k, host, port)
#
# Where host and port configure the network address on which the monitor
# operates.
#
# To connect to the monitor, run python3 -m curio.monitor [port]. For example:
#
# Theory of operation:
# --------------------
# The monitor works by opening up a loopback socket on the local machine and
# allowing connections via telnet. By default, it only allows a connection
# originating from the local machine.  Only a single monitor connection is
# allowed at any given time.
#
# There are two parts to the monitor itself: a user interface and an
# internal loop that runs on curio itself.  The user interface part
# runs in a completely separate execution thread.  The reason for this
# is that it allows curio to be monitored even if the curio kernel is
# completely deadlocked, occupied with a large CPU-bound task, or
# otherwise hosed in the some way.  At a minimum, you can connect,
# look at the task table, and see what the tasks are doing.
#
# The internal monitor loop implemented on curio itself is presently
# used to implement external task cancellation.  Manipulating any 
# part of the kernel state or task status is unsafe from an outside thread.
# To make it safe, the user-interface thread of the monitor hands over
# requests requiring the involvement of the kernel to the monitor loop.
# Since this loop runs on curio, it can safely make cancellation requests
# and perform other kernel-related actions. 

import sys
import os
import traceback
import linecache
import signal
import time
import socket
import threading
import queue
import telnetlib
from collections import deque
import logging

# --- Curio
from .task import Task

# ---
log = logging.getLogger(__name__)

MONITOR_HOST = '127.0.0.1'
MONITOR_PORT = 48802

def _get_stack(task):
    '''
    Extracts a list of stack frames from a chain of generator/coroutine calls
    '''
    frames = []
    coro = task.coro
    while coro:
        f = coro.cr_frame if hasattr(coro, 'cr_frame') else coro.gi_frame
        if f is not None:
            frames.append(f)
        coro = coro.cr_await if hasattr(coro, 'cr_await') else coro.gi_yieldfrom
    return frames

def _format_stack(task):
    '''
    Formats a traceback from a stack of coroutines/generators
    '''
    extracted_list = []
    checked = set()
    for f in _get_stack(task):
        lineno = f.f_lineno
        co = f.f_code
        filename = co.co_filename
        name = co.co_name
        if filename not in checked:
            checked.add(filename)
            linecache.checkcache(filename)
        line = linecache.getline(filename, lineno, f.f_globals)
        extracted_list.append((filename, lineno, name, line))
    if not extracted_list:
        resp = 'No stack for %r' % task 
    else:
        resp = 'Stack for %r (most recent call last):\n' % task
        resp += ''.join(traceback.format_list(extracted_list))
    return resp

class Monitor(object):
    '''
    Task monitor that runs concurrently to the curio kernel in a 
    separate thread. This can watch the kernel and provide debugging.
    '''
    def __init__(self, kern, host=MONITOR_HOST, port=MONITOR_PORT):
        self.kernel = kern
        self.address = (host, port)
        self.monitor_queue = queue.Queue()

        log.info('Starting Curio monitor at %s:%d', host, port)

        # The monitor launches both a separate thread and helper task
        # that runs inside curio itself to manage cancellation events
        threading.Thread(target=self.server, daemon=True).start()
        monitor_task = Task(self.monitor_task(), daemon=True)
        kern._ready.append(monitor_task)
        kern._tasks[monitor_task.id] = monitor_task

    async def monitor_task(self):
        '''
        Asynchronous task loop for carrying out task cancellation.
        '''
        from .sync import abide
        while True:
            task = await abide(self.monitor_queue.get)
            await task.cancel()

    def server(self):
        '''
        Synchronous kernel for the monitor.  This runs in a separate thread
        from curio itself.
        '''
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
        sock.bind(self.address)
        sock.listen(1)
        while True:
            client, addr = sock.accept()
            sout = client.makefile('w', encoding='utf-8')
            sin = client.makefile('r', encoding='utf-8')
            self.interactive_loop(sout, sin)
            client.close()
                
    def interactive_loop(self, sout, sin):
        '''
        Main interactive loop of the monitor
        '''
        sout.write('\nCurio Monitor: %d tasks running\n' % len(self.kernel._tasks))
        sout.write('Type help for commands\n')
        while True:
            sout.write('curio > ')
            sout.flush()
            resp = sin.readline()
            try:
                if not resp or resp.startswith('q'):
                    sout.write('Leaving monitor\n')
                    return

                elif resp.startswith('p'):
                    self.command_ps(sout)

                elif resp.startswith('exit'):
                    self.command_exit(sout)

                elif resp.startswith('cancel'):
                    _, taskid_s = resp.split()
                    self.command_cancel(sout, int(taskid_s))

                elif resp.startswith('signal'):
                    _, signame = resp.split()
                    self.command_signal(sout, signame)

                elif resp.startswith('w'):
                    _, taskid_s = resp.split()
                    self.command_where(sout, int(taskid_s))

                elif resp.startswith('h'):
                    self.command_help(sout)
                else:
                    sout.write('Unknown command. Type help.\n')
            except Exception as e:
                sout.write('Bad command. %s\n' % e)

    def command_help(self, sout):
        sout.write(
     '''Commands:
         ps               : Show task table
         where taskid     : Show stack frames for a task
         cancel taskid    : Cancel an indicated task
         signal signame   : Send a Unix signal
         quit             : Leave the monitor
''')

    def command_ps(self, sout):
        headers = ('Task', 'State', 'Cycles', 'Timeout', 'Task')
        widths = (6, 12, 10, 7, 50)
        for h, w in zip(headers, widths):
            sout.write('%-*s ' % (w, h))
        sout.write('\n')
        sout.write(' '.join(w*'-' for w in widths))
        sout.write('\n')
        timestamp = time.monotonic()
        for taskid in sorted(self.kernel._tasks):
            task = self.kernel._tasks.get(taskid)
            if task:
                remaining = format((task.timeout - timestamp), '0.6f')[:7] if task.timeout else 'None'
                sout.write('%-*d %-*s %-*d %-*s %-*s\n' % (widths[0], taskid, 
                                                                widths[1], task.state,
                                                                widths[2], task.cycles,
                                                                widths[3], remaining,
                                                                widths[4], task))

    def command_where(self, sout, taskid):
        task = self.kernel._tasks.get(taskid)
        if task:
            stack = _format_stack(task)
            sout.write(stack+'\n')
        else:
            sout.write('No task %d\n' % taskid)

    def command_signal(self, sout, signame):
        if hasattr(signal, signame):
            os.kill(os.getpid(), getattr(signal, signame))
        else:
            sout.write('Unknown signal %s\n' % signame)

    def command_cancel(self, sout, taskid):
        task = self.kernel._tasks.get(taskid)
        if task:
            sout.write('Cancelling task %d\n' % taskid)
            self.monitor_queue.put(task)
            
    def command_exit(self, sout):
        pass

def monitor_client(host, port):
    '''
    Client to connect to the monitor via "telnet"
    '''
    tn = telnetlib.Telnet()
    tn.open(host, port, timeout=0.5)
    tn.interact()
    tn.close()

if __name__ == '__main__':
    if sys.argv[1:]:
        port = int(sys.argv[1])
    else:
        port = MONITOR_PORT
    monitor_client(MONITOR_HOST, port)



# curio/meta.py
#     ___        
#     \./      DANGER:  This module implements some experimental
#  .--.O.--.            metaprogramming techniques involving async/await.
#   \/   \/             If you use it, you might die. 
#

__all__ = [ 'acontextmanager' ]

import sys
import inspect
from functools import wraps
from contextlib import contextmanager

class acontextmanager(object):
    '''
    Decorator that allows an asynchronous context manager to be written
    using the same technique as with the @contextlib.contextmanager 
    decorator.
    '''
    def __init__(self, func):
        self.manager = contextmanager(func)

    def __aenter__(self):
        return self.manager.__enter__()

    def __aexit__(self, *args):
        return self.manager.__exit(*args)

    def __enter__(self):
        raise RuntimeError('Use async with')

    def __exit__(self, *args):
        pass

def awaitable(syncfunc):
    '''
    Decorator that allows an asynchronous function to be paired with a 
    synchronous function in a single function call.  The selection of
    which function executes depends on the calling context.  For example:

        def spam(sock, maxbytes):                       (A)
            return sock.recv(maxbytes)

        @awaitable(read)                                (B)
        async def spam(sock, maxbytes):
            return await sock.recv(maxbytes)

    In later code, you could use the spam() function in either a synchronous
    or asynchronous context.  For example:

        def foo():
            ...
            r = spam(s, 1024)          # Calls synchronous function (A) above
            ... 
    
        async def bar():
            ...
            r = await spam(s, 1024)    # Calls async function (B) above
            ...

    '''
    def decorate(asyncfunc):
        if inspect.signature(syncfunc) != inspect.signature(asyncfunc):
            raise TypeError('%s and async %s have different signatures' %
                            (syncfunc.__name__, asyncfunc.__name__))
        @wraps(asyncfunc)
        def wrapper(*args, **kwargs):
            if sys._getframe(1).f_code.co_flags & 0x80:
                return asyncfunc(*args, **kwargs)
            else:
                return syncfunc(*args, **kwargs)
        return wrapper
    return decorate









# test_channel.py

import pytest
from socket import *
from curio.channel import Channel
from curio.io import SocketStream
from curio import spawn, sleep, CancelledError, TaskTimeout, timeout_after

@pytest.fixture
def chs():
    sock1, sock2 = socketpair()
    sock1_s = SocketStream(sock1)
    sock2_s = SocketStream(sock2)
    ch1 = Channel(sock1_s, sock1_s)
    ch2 = Channel(sock2_s, sock2_s)
    return (ch1, ch2)

    sock1, sock2 = socketpair()
    fileno1 = sock1.detach()
    ch1 = Channel(Stream(open(fileno1, 'rb', buffering=0)),
                  Stream(open(fileno1, 'wb', buffering=0, closefd=False)))

    fileno2 = sock2.detach()
    ch2 = Channel(Stream(open(fileno2, 'rb', buffering=0)),
                  Stream(open(fileno2, 'wb', buffering=0, closefd=False)))
    return (ch1, ch2)

def test_channel_hello(kernel, chs):
    results = []

    async def server(ch):
        async with ch:
            await ch.send('server hello world')
            results.append(await ch.recv())

    async def client(ch):
        async with ch:
            msg = await ch.recv()
            results.append(msg)
            await ch.send('client hello world')
        
    async def main(ch1, ch2):
         await spawn(server(ch1))
         await spawn(client(ch2))

    kernel.run(main(*chs))
    assert results == [ 'server hello world',
                        'client hello world' ]

def test_channel_hello_bytes(kernel, chs):
    results = []

    async def server(ch):
        async with ch:
            await ch.send(b'server hello world')
            results.append(await ch.recv())

    async def client(ch):
        async with ch:
            msg = await ch.recv()
            results.append(msg)
            await ch.send(b'client hello world')

    async def main(ch1, ch2):
         await spawn(server(ch1))
         await spawn(client(ch2))

    kernel.run(main(*chs))
    assert results == [ b'server hello world',
                        b'client hello world' ]

def test_channel_large(kernel, chs):
    results = []
    data = list(range(1000000))

    async def server(ch):
        async with ch:
            await ch.send(data)
            results.append(await ch.recv())

    async def client(ch):
        async with ch:
            msg = await ch.recv()
            results.append(msg)
            await ch.send(len(msg))

    async def main(ch1, ch2):
         await spawn(server(ch1))
         await spawn(client(ch2))

    kernel.run(main(*chs))
    assert results == [ data,
                        len(data) ]

def test_channel_auth(kernel, chs):
    results = []

    async def server(ch):
        async with ch:
            await ch.authenticate_server(b'peekaboo')
            await ch.send('server hello world')
            results.append(await ch.recv())

    async def client(ch):
        async with ch:
            await ch.authenticate_client(b'peekaboo')
            msg = await ch.recv()
            results.append(msg)
            await ch.send('client hello world')

    async def main(ch1, ch2):
         await spawn(server(ch1))
         await spawn(client(ch2))

    kernel.run(main(*chs))

    assert results == [ 'server hello world',
                        'client hello world' ]

def test_channel_send_partial_bytes(kernel, chs):
    results = []
    data = b'abcdefghijklmnopqrstuvwxyz'

    async def server(ch):
        async with ch:
            await ch.send_bytes(data, offset=5, size=10)
            results.append(await ch.recv())
            await ch.send_bytes(data, offset=5)
            results.append(await ch.recv())
            await ch.send_bytes(data, size=10)
            results.append(await ch.recv())

            # Try some bad inputs
            try:
                await ch.send_bytes(data, offset=50)
            except ValueError as e:
                results.append(str(e))

            try:
                await ch.send_bytes(data, size=50)
            except ValueError as e:
                results.append(str(e))

            try:
                await ch.send_bytes(data, offset=-10)
            except ValueError as e:
                results.append(str(e))

            try:
                await ch.send_bytes(data, size=-10)
            except ValueError as e:
                results.append(str(e))

    async def client(ch):
        async with ch:
            msg = await ch.recv_bytes()
            results.append(msg)
            await ch.send(len(msg))
            msg = await ch.recv_bytes()
            results.append(msg)
            await ch.send(len(msg))
            msg = await ch.recv_bytes()
            results.append(msg)
            await ch.send(len(msg))

    async def main(ch1, ch2):
         await spawn(server(ch1))
         await spawn(client(ch2))

    kernel.run(main(*chs))
    assert results == [ data[5:15], 10,
                        data[5:], len(data[5:]),
                        data[:10], 10,
                        'buffer length < offset',
                        'buffer length < offset + size',
                        'offset is negative',
                        'size is negative',

                        ]

def test_channel_from_connection(kernel):
    import multiprocessing
    p1, p2 = multiprocessing.Pipe()
    ch1 = Channel.from_Connection(p1)
    ch2 = Channel.from_Connection(p2)
    
    results = []

    async def server(ch):
        async with ch:
            await ch.send('server hello world')
            results.append(await ch.recv())

    async def client(ch):
        async with ch:
            msg = await ch.recv()
            results.append(msg)
            await ch.send('client hello world')
        
    async def main(ch1, ch2):
         await spawn(server(ch1))
         await spawn(client(ch2))

    kernel.run(main(ch1, ch2))
    assert results == [ 'server hello world',
                        'client hello world' ]


def test_channel_recv_cancel(kernel, chs):
    results = []

    async def client(ch):
        async with ch:
            try:
                msg = await ch.recv()
                results.append(msg)
            except CancelledError:
                results.append('cancel')

    async def main(ch):
        task = await spawn(client(ch))
        await sleep(1)
        await task.cancel()
        results.append('done cancel')

    ch1, ch2 = chs
    kernel.run(main(ch2))
    assert results == [ 'cancel', 'done cancel' ]


def test_channel_recv_timeout(kernel, chs):
    results = []

    async def client(ch):
        try:
            msg = await timeout_after(1.0, ch.recv())
            results.append(msg)
        except TaskTimeout:
            results.append('timeout')

    async def main(ch):
        task = await spawn(client(ch))
        await task.join()
        results.append('done')

    ch1, ch2 = chs
    kernel.run(main(ch2))
    assert results == [ 'timeout', 'done' ]

def test_channel_send_cancel(kernel, chs):
    results = []

    async def client(ch):
        async with ch:
            try:
                msg = 'x'*10000000   # Should be large enough to cause send blocking
                await ch.send(msg)
                results.append('success')
            except CancelledError:
                results.append('cancel')

    async def main(ch):
        task = await spawn(client(ch))
        await sleep(1)
        await task.cancel()
        results.append('done cancel')

    ch1, ch2 = chs
    kernel.run(main(ch2))
    assert results == [ 'cancel', 'done cancel' ]


def test_channel_send_timeout(kernel, chs):
    results = []

    async def client(ch):
        try:
            msg = 'x'*10000000
            await timeout_after(1, ch.send(msg))
            results.append('success')
        except TaskTimeout:
            results.append('timeout')

    async def main(ch):
        task = await spawn(client(ch))
        await task.join()
        results.append('done')

    ch1, ch2 = chs
    kernel.run(main(ch2))
    assert results == [ 'timeout', 'done' ]



# curio/kernel.py
#
# Main execution kernel.

import socket
import heapq
import time
import os
import sys
import logging
import signal
from selectors import DefaultSelector, EVENT_READ, EVENT_WRITE
from collections import deque, defaultdict

# Logger where uncaught exceptions from crashed tasks are logged
log = logging.getLogger(__name__)

from .errors import *
from .errors import _CancelRetry
from .task import Task
from .traps import _read_wait

# kqueue is the datatype used by the kernel for all of its queuing functionality.
# Any time a task queue is needed, use this type instead of directly hard-coding the
# use of a deque.  This will make sure the code continues to work even if the
# queue type is changed later.

kqueue = deque

# ----------------------------------------------------------------------
# Underlying kernel that drives everything
# ----------------------------------------------------------------------

class Kernel(object):
    def __init__(self, *, selector=None, with_monitor=False, pdb=False, log_errors=True):
        if selector is None:
            selector = DefaultSelector()
        self._selector = selector
        self._ready = kqueue()            # Tasks ready to run
        self._tasks = { }                 # Task table
        self._signal_sets = None          # Dict { signo: [ sigsets ] } of watched signals (initialized only if signals used)
        self._default_signals = None      # Dict of default signal handlers

        # Attributes related to the loopback socket (only initialized if required)
        self._notify_sock = None
        self._wait_sock = None
        self._kernel_task_id = None

        # Wake queue for tasks restarted by external threads
        self._wake_queue = deque()

        # Sleeping task queue
        self._sleeping = []

        # Optional process/thread pools (see workers.py)
        self._thread_pool = None
        self._process_pool = None

        # Optional crash handler callback
        self._crash_handler = None

        self._pdb = pdb
        self._log_errors = log_errors

        # If a monitor is specified, launch it
        if with_monitor or 'CURIOMONITOR' in os.environ:
            Monitor(self)

    def __del__(self):
        if self._kernel_task_id:
            self._notify_sock.close()
            self._wait_sock.close()
            self._kernel_task_id = None
        if self._thread_pool:
            self._thread_pool.shutdown()
        if self._process_pool:
            self._process_pool.shutdown()

    # Main Kernel Loop
    # ----------
    def run(self, coro=None, *, shutdown=False):
        '''
        Run the kernel until no more non-daemonic tasks remain.  If
        shutdown is True, the kernel cleans up after itself after all
        tasks complete.
        '''
        
        # Motto:  "What happens in the kernel stays in the kernel"

        # ---- Kernel State
        current = None                          # Currently running task
        selector = self._selector               # Event selector
        ready = self._ready                     # Ready queue
        tasks = self._tasks                     # Task table
        sleeping = self._sleeping               # Sleeping task queue
        wake_queue = self._wake_queue           # External wake queue

        # ---- Number of non-daemonic tasks running
        njobs = sum(not task.daemon for task in tasks.values())
        
        # ---- Bound methods
        selector_register = selector.register
        selector_unregister = selector.unregister
        selector_select = selector.select
        ready_popleft = ready.popleft
        ready_append = ready.append
        ready_appendleft = ready.appendleft
        time_monotonic = time.monotonic

        # ---- In-kernel task used for processing signals and futures

        # Force the kernel to wake, possibly scheduling a task to run.
        # This method is called by threads running concurrently to the
        # curio kernel.  For example, it's triggered upon completion of
        # Futures created by thread pools and processes. It's inherently
        # dangerous for any kind of operation on the kernel to be
        # performed by a separate thread.  Thus, the *only* thing that
        # happens here is that the task gets appended to a deque and a
        # notification message is written to the kernel notification
        # socket.  append() and pop() operations on deques are thread safe
        # and do not need additional locking.  See
        # https://docs.python.org/3/library/collections.html#collections.deque
        # ----------
        def _wake(task=None, future=None):
            if task:
                wake_queue.append((task, future))
            self._notify_sock.send(b'\x00')

        # Initialize the loopback socket and launch the kernel task if needed
        def _init_loopback():
            self._notify_sock, self._wait_sock = socket.socketpair()
            self._wait_sock.setblocking(False)
            self._notify_sock.setblocking(False)
            task = Task(_kernel_task(), taskid=0)
            _reschedule_task(task)
            self._kernel_task_id = task.id

        # Internal task that monitors the loopback socket--allowing the kernel to
        # awake for non-I/O events.  Also processes incoming signals.  This only
        # launches if needed to wait for external events (futures, signals, etc.)
        async def _kernel_task():
            wake_queue_popleft = wake_queue.popleft
            wait_sock = self._wait_sock

            while True:
                await _read_wait(wait_sock)
                data = wait_sock.recv(1000)

                # Process any waking tasks.  These are tasks that have been awakened
                # externally to the event loop (e.g., by separate threads, Futures, etc.)
                while wake_queue:
                    task, future = wake_queue_popleft()
                    # If the future associated with wakeup no longer matches
                    # the future stored on the task, wakeup is abandoned.
                    # It means that a timeout or cancellation event occurred
                    # in the time interval between the call to self._wake()
                    # and the subsequent processing of the waking task
                    if future and task.future is not future:
                        continue
                    task.future = None
                    task.state = 'READY'
                    task.cancel_func = None
                    ready_append(task)

                # Any non-null bytes received here are assumed to be received signals.
                # See if there are any pending signal sets and unblock if needed 
                if not self._signal_sets:
                    continue

                sigs = (signal.Signals(n) for n in data if n in self._signal_sets)
                for signo in sigs:
                    for sigset in self._signal_sets[signo]:
                        sigset.pending.append(signo)
                        if sigset.waiting:
                            _reschedule_task(sigset.waiting, value=signo)
                            sigset.waiting = None

        # ---- Task Support Functions

        # Create a new task. Putting it on the ready queue
        def _new_task(coro, daemon=False):
            nonlocal njobs
            task = Task(coro, daemon)
            tasks[task.id] = task
            if not daemon:
                njobs += 1
            _reschedule_task(task)
            return task

        # Reschedule a task, putting it back on the ready queue so that it can run.
        # value and exc specify a value or exception to send into the underlying 
        # coroutine when it is rescheduled.
        def _reschedule_task(task, value=None, exc=None):
            ready_append(task)
            task.next_value = value
            task.next_exc = exc
            task.state = 'READY'
            task.cancel_func = None

        # Cancel a task. This causes a CancelledError exception to raise in the
        # underlying coroutine.  The coroutine can elect to catch the exception
        # and continue to run to perform cleanup actions.  However, it would 
        # normally terminate shortly afterwards.   This method is also used to 
        # raise timeouts.
        def _cancel_task(task, exc=CancelledError):
            if task.terminated:
                return True

            if not task.cancel_func:
                # All tasks set a cancellation function when they
                # block.  If there is no cancellation function set. It
                # means that the task finished whatever it might have
                # been working on and it's sitting in the ready queue
                # ready to run.  This presents a rather tricky corner
                # case of cancellation. First, it means that the task
                # successfully completed whatever it was waiting to
                # do, but it has not yet communicated the result back.
                # This could be something critical like acquiring a
                # lock.  Because of that, can't just arbitrarily nuke
                # the task.  Instead, we have to let it be rescheduled
                # and properly process the result of the successfully
                # completed operation.  The return of False here means
                # that cancellation failed and that it should be
                # retried.  Public facing API calls that cancel tasks
                # need to check this return value and retry.
                return False

            # Detach the task from where it might be waiting at this moment
            task.cancel_func()

            # Reschedule it with a pending exception
            _reschedule_task(task, exc=exc(exc.__name__))
            return True

        # Cleanup task.  This is called after the underlying coroutine has
        # terminated.  value and exc give the return value or exception of
        # the coroutine.  This wakes any tasks waiting to join.
        def _cleanup_task(task, value=None, exc=None):
            nonlocal njobs
            task.next_value = value
            task.next_exc = exc
            task.timeout = None

            if not task.daemon:
                njobs -=1

            if task.joining:
                for wtask in task.joining:
                    _reschedule_task(wtask) 
                task.joining = None
            task.terminated = True

            del tasks[task.id]

        # Shut down the kernel, cleaning up resources and cancelling
        # still-running daemon tasks
        def _shutdown():
            nonlocal njobs
            for task in sorted(tasks.values(), key=lambda t: t.id, reverse=True):
                if task.id == self._kernel_task_id:
                    continue

                # If the task is daemonic, force it to non-daemon status and cancel it
                if task.daemon:
                    njobs += 1
                    task.daemon = False

                assert _cancel_task(task)

            # Run all of the daemon tasks through cancellation
            if ready:
                self.run()

            # Cancel the kernel loopback task (if any)
            task = tasks.pop(self._kernel_task_id, None)
            if task:
                task.cancel_func()
                self._notify_sock.close()
                self._notify_sock = None
                self._wait_sock.close()
                self._wait_sock = None
                self._kernel_task_id = None

            # Remove the signal handling file descriptor (if any)
            if self._signal_sets:
                signal.set_wakeup_fd(-1)
                self._signal_sets = None
                self._default_signals = None

        # Set a timeout or sleep event on the current task
        def _set_timeout(seconds, sleep_type='timeout'):
            timeout = time_monotonic() + seconds
            item = (timeout, current.id, sleep_type)
            heapq.heappush(sleeping, item)
            setattr(current, sleep_type, timeout)

        # ---- Traps
        #
        # These implement the low-level functionality that is
        # triggered by coroutines.  They are never invoked directly
        # and there is no public API outside the kernel.  Instead,
        # coroutines use a statement such as
        #   
        #   yield ('_trap_io', sock, EVENT_READ, 'READ_WAIT')
        #
        # to invoke a specific trap.

        # Wait for I/O
        def _trap_io(_, fileobj, event, state):
            # See comment about deferred unregister in run().  If the requested
            # I/O operation is *different* than the last I/O operation that was
            # performed by the task, we need to unregister the last I/O resource used
            # and register a new one with the selector.  
            if current._last_io != (state, fileobj):
                if current._last_io:
                    selector_unregister(current._last_io[1])
                selector_register(fileobj, event, current)

            # This step indicates that we have managed any deferred I/O management
            # for the task.  Otherwise the run() method will perform an unregistration step.
            current._last_io = None
            current.state = state
            current.cancel_func = lambda: selector_unregister(fileobj)

        # Wait on a Future
        def _trap_future_wait(_, future, event):
            if not self._kernel_task_id:
                _init_loopback()

            current.state = 'FUTURE_WAIT'
            current.cancel_func = (lambda task=current: 
                                   setattr(task, 'future', future.cancel() and None))
            current.future = future

            # Discussion: Each task records the future that it is
            # currently waiting on.  The completion callback below only
            # attempts to wake the task if its stored Future is exactly
            # the same one that was stored above.  Due to support for
            # cancellation and timeouts, it's possible that a task might
            # abandon its attempt to wait for a Future and go on to
            # perform other operations, including waiting for different
            # Future in the future (got it?).  However, a running thread
            # or process still might go on to eventually complete the
            # earlier work.  In that case, it will trigger the callback,
            # find that the task's current Future is now different, and
            # discard the result.
            future.add_done_callback(lambda fut, task=current: 
                                     _wake(task, fut) if task.future is fut else None)

            # An optional threading.Event object can be passed and set to
            # start a worker thread.   This makes it possible to have a lock-free
            # Future implementation where worker threads only start after the
            # callback function has been set above.
            if event:
                event.set()

        # Add a new task to the kernel
        def _trap_spawn(_, coro, daemon):
            task = _new_task(coro, daemon)
            _reschedule_task(current, value=task)

        # Reschedule one or more tasks from a queue
        def _trap_reschedule_tasks(_, queue, n, value, exc):
            while n > 0:
                _reschedule_task(queue.popleft(), value=value, exc=exc)
                n -= 1
            _reschedule_task(current)

        # Join with a task
        def _trap_join_task(_, task):
            if task.terminated:
                _reschedule_task(current) 
            else:
                if task.joining is None:
                    task.joining = kqueue()
                _trap_wait_queue(_, task.joining, 'TASK_JOIN')

        # Cancel a task
        def _trap_cancel_task(_, task):
            if task == current:
                _reschedule_task(current, exc=CurioError("A task can't cancel itself"))
                return

            if task.cancelled or _cancel_task(task, CancelledError):
                task.cancelled = True
                _trap_join_task(_, task)

            else:
                # Fail with a _CancelRetry exception to indicate that the cancel
                # request should be attempted again.  This happens in the case
                # that a cancellation request is issued against a task that
                # ready to run in the ready queue.
                _reschedule_task(current, exc=_CancelRetry())

        # Wait on a queue
        def _trap_wait_queue(_, queue, state):
            queue.append(current)
            current.state = state
            current.cancel_func = lambda current=current: queue.remove(current)

        # Sleep for a specified period
        def _trap_sleep(_, seconds):
            if seconds > 0:
                _set_timeout(seconds, 'sleep')
                current.state = 'TIME_SLEEP'
                current.cancel_func = lambda task=current: setattr(task, 'sleep', None)
            else:
                _reschedule_task(current)

        # Watch signals
        def _trap_sigwatch(_, sigset):
            # Initialize the signal handling part of the kernel if not done already
            # Note: This only works if running in the main thread
            if self._signal_sets is None:
                self._signal_sets = defaultdict(list)
                self._default_signals = { }
                if not self._kernel_task_id:
                    _init_loopback()
                old_fd = signal.set_wakeup_fd(self._notify_sock.fileno())     
                assert old_fd < 0, 'Signals already initialized %d' % old_fd

            for signo in sigset.signos:
                if not self._signal_sets[signo]:
                    self._default_signals[signo] = signal.signal(signo, lambda signo, frame:None)
                self._signal_sets[signo].append(sigset)

            _reschedule_task(current)

        # Unwatch signals
        def _trap_sigunwatch(_, sigset):
            for signo in sigset.signos:
                if sigset in self._signal_sets[signo]:
                    self._signal_sets[signo].remove(sigset)

                # If there are no active watchers for a signal, revert it back to default behavior
                if not self._signal_sets[signo]:
                    signal.signal(signo, self._default_signals[signo])
                    del self._signal_sets[signo]
            _reschedule_task(current)

        # Wait for a signal
        def _trap_sigwait(_, sigset):
            sigset.waiting = current
            current.state = 'SIGNAL_WAIT'
            current.cancel_func = lambda: setattr(sigset, 'waiting', None)

        # Set a timeout to be delivered to the calling task
        def _trap_set_timeout(_, seconds):
            old_timeout = current.timeout
            if seconds:
                _set_timeout(seconds)
            else:
                current.timeout = None
            current.next_value = old_timeout
            ready_appendleft(current)

        # Clear a previously set timeout
        def _trap_unset_timeout(_, previous):
            current.timeout = previous
            current.next_value = None
            ready_appendleft(current)

        # Return the running kernel
        def _trap_get_kernel(_):
            ready_appendleft(current)
            current.next_value = self

        # Return the currently running task
        def _trap_get_current(_):
            ready_appendleft(current)
            current.next_value = current

        # Create the traps table
        traps = { name: trap 
                  for name, trap in locals().items()
                  if name.startswith('_trap_') }
        
        # If a coroutine was given, add it as the first task
        maintask = _new_task(coro) if coro else None

        # ------------------------------------------------------------
        # Main Kernel Loop
        # ------------------------------------------------------------
        while njobs > 0:
            # --------
            # Poll for I/O as long as there is nothing to run
            # --------
            while not ready:
                # Wait for an I/O event (or timeout)
                timeout = (sleeping[0][0] - time_monotonic()) if sleeping else None
                events = selector_select(timeout)

                # Reschedule tasks with completed I/O
                for key, mask in events:
                    task = key.data
                    # Please see comment regarding deferred_unregister in run()
                    task._last_io = (task.state, key.fileobj)
                    task.state = 'READY'
                    task.cancel_func = None
                    ready_append(task)

                # Process sleeping tasks (if any)
                if sleeping:
                    current_time = time_monotonic()
                    while sleeping and sleeping[0][0] <= current_time:
                        tm, taskid, sleep_type = heapq.heappop(sleeping)
                        # When a task wakes, verify that the timeout value matches that stored
                        # on the task. If it differs, it means that the task completed its
                        # operation, was cancelled, or is no longer concerned with this
                        # sleep operation.  In that case, we do nothing
                        if taskid in tasks:
                            task = tasks[taskid]
                            if sleep_type == 'sleep':
                                if tm == task.sleep:
                                    task.sleep = None
                                    _reschedule_task(task)
                            else:
                                if tm == task.timeout:
                                    task.timeout = None
                                    _cancel_task(task, exc=TaskTimeout)

            # --------
            # Run ready tasks
            # --------
            while ready:
                current = ready_popleft()
                try:
                    current.state = 'RUNNING'
                    current.cycles += 1
                    if current.next_exc is None:
                        trap = current._send(current.next_value)
                    else:
                        trap = current._throw(current.next_exc)
                        current.next_exc = None

                    # Execute the trap
                    traps[trap[0]](*trap) 

                except StopIteration as e:
                    _cleanup_task(current, value = e.value)

                except CancelledError as e:
                    current.exc_info = sys.exc_info()
                    current.state = 'CANCELLED'
                    _cleanup_task(current)

                except Exception as e:
                    current.exc_info = sys.exc_info()
                    current.state = 'CRASHED'
                    exc = TaskError('Task Crashed')
                    exc.__cause__ = e
                    _cleanup_task(current, exc=exc)
                    if self._log_errors:
                        log.error('Curio: Task Crash: %s' % current, exc_info=True)

                    if self._crash_handler:
                        self._crash_handler(current)

                    if self._pdb:
                        import pdb as _pdb
                        _pdb.post_mortem(current.exc_info[2])

                except SystemExit:
                    _cleanup_task(current)
                    raise

                finally:
                    # Unregister previous I/O request. Discussion follows:
                    # 
                    # When a task performs I/O, it registers itself with the underlying
                    # I/O selector.  When the task is reawakened, it unregisters itself
                    # and prepares to run.  However, in many network applications, the
                    # task will perform a small amount of work and then go to sleep on
                    # exactly the same I/O resource that it was waiting on before. For
                    # example, a client handling task in a server will often spend most
                    # of its time waiting for incoming data on a single socket. 
                    #
                    # Instead of always unregistering the task from the selector, we
                    # can defer the unregistration process until after the task goes
                    # back to sleep again.  If it happens to be sleeping on the same
                    # resource as before, there's no need to unregister it--it will
                    # still be registered from the last I/O operation.   
                    #
                    # The code here performs the unregister step for a task that 
                    # ran, but is now sleeping for a *different* reason than repeating the
                    # prior I/O operation.  There is coordination with code in _trap_io().

                    if current._last_io:
                        selector_unregister(current._last_io[1])
                        current._last_io = None
                    
        # If kernel shutdown has been requested, issue a cancellation request to all remaining tasks
        if shutdown:
            _shutdown()

        return maintask.next_value if maintask else None

def run(coro, *, pdb=False, log_errors=True, with_monitor=False, selector=None):
    '''
    Run the curio kernel with an initial task and execute until all
    tasks terminate.  Returns the task's final result (if any). This
    is a convenience function that should primarily be used for
    launching the top-level task of an curio-based application.  It
    creates an entirely new kernel, runs the given task to completion,
    and concludes by shutting down the kernel, releasing all resources used.
    
    Don't use this function if you're repeatedly launching a lot of
    new tasks to run in curio. Instead, create a Kernel instance and
    use its run() method instead.
    '''
    kernel = Kernel(selector=selector, with_monitor=with_monitor, 
                    log_errors=log_errors, pdb=pdb)
    result = kernel.run(coro, shutdown=True)
    return result

__all__ = [ 'Kernel', 'run' ]
            
from .monitor import Monitor
        



# curio/io.py
#
# Curio is primarily concerned with the scheduling of tasks. In
# particular, the kernel does not actually perform any I/O.  It merely
# blocks tasks that need to wait for reading or writing.  To actually
# perform I/O, you use the existing file and socket abstractions
# already provided by the Python standard library.  The only
# difference is that you need to take extra steps to manage their
# non-blocking behavior.  The classes in this file provide wrappers
# around socket-like and file-like objects. Methods responsible for
# reading/writing have a small amount of extra logic to added to
# handle their scheduling.  Other methods are simply passed through to
# the original object via delegation.
#
# It's important to emphasize that these classes can be applied to
# *ANY* existing socket-like or file-like object as long as it
# represents a real system-level file (must have a fileno() method)
# and can be used with the underlying I/O selector.  For example, the
# Socket class can wrap a normal socket or an SSL socket--it doesn't
# matter which.  Similarly, the Stream class can wrap normal files,
# files created from sockets, pipes, and other file-like abstractions.
#
# No assumption is made about system compatibility (Unix vs. Windows).  
# The main compatibility concern would be at the level of the I/O
# selector used by the kernel.  For example, can it detect I/O events
# on the provided file or socket?  If so, it will probably work here.

__all__ = ['Socket', 'FileStream', 'SocketStream']

from socket import SOL_SOCKET, SO_ERROR
from contextlib import contextmanager
import io
import os

from .traps import _read_wait, _write_wait

# Exceptions raised for non-blocking I/O.  For normal sockets, blocking operations
# normally just raise BlockingIOError.  For SSL sockets, more specific exceptions
# are raised.  Here we're just making some aliases for the possible exceptions.

try:
    from ssl import SSLWantReadError, SSLWantWriteError
    WantRead = (BlockingIOError, SSLWantReadError)
    WantWrite = (BlockingIOError, SSLWantWriteError)
except ImportError:
    WantRead = BlockingIOError
    WantWrite = BlockingIOError

# There is a certain amount of repetition in this class.  It can
# probably be shortened with some sort of decorator magic. On the
# other, the KISSS (Keep it Stupid Simple Stupid) principle might be a
# better policy--just in case someone needs to debug it.

class Socket(object):
    '''
    Non-blocking wrapper around a socket object.   The original socket is put
    into a non-blocking mode when it's wrapped.
    '''
    def __init__(self, sock):
        self._socket = sock
        self._socket.setblocking(False)
        self._fileno = sock.fileno()

        # Commonly used bound methods
        self._socket_send = sock.send
        self._socket_recv = sock.recv

    def __repr__(self):
        return '<curio.Socket %r>' % (self._socket)

    def __getattr__(self, name):
        return getattr(self._socket, name)
            
    def fileno(self):
        return self._fileno

    def settimeout(self, seconds):
        raise RuntimeError('Use timeout_after() to set a timeout')

    def gettimeout(self):
        return None

    def dup(self):
        return Socket(self._socket.dup())

    def makefile(self, mode, buffering=0, *, encoding=None, errors=None, newline=None):
        if 'b' not in mode:
            raise RuntimeError('File can only be created in binary mode')
        f = self._socket.makefile(mode, buffering=buffering)
        return FileStream(f)

    def as_stream(self):
        '''
        Create a stream-based interface to the socket.
        '''
        return SocketStream(self._socket)

    @contextmanager
    def blocking(self):
        '''
        Allow temporary access to the underlying socket in blocking mode
        '''
        try:
            self._socket.setblocking(True)
            yield self._socket
        finally:
            self._socket.setblocking(False)

    async def recv(self, maxsize, flags=0):
        while True:
            try:
                return self._socket_recv(maxsize, flags)
            except WantRead:
                await _read_wait(self._fileno)
            except WantWrite:
                await _write_wait(self._fileno)

    async def recv_into(self, buffer, nbytes=0, flags=0):
        while True:
            try:
                return self._socket.recv_into(buffer, nbytes, flags)
            except WantRead:
                await _read_wait(self._fileno)
            except WantWrite:
                await _write_wait(self._fileno)

    async def send(self, data, flags=0):
        while True:
            try:
                return self._socket_send(data, flags)
            except WantWrite:
                await _write_wait(self._fileno)
            except WantRead:
                await _read_wait(self._fileno)

    async def sendall(self, data, flags=0):
        buffer = memoryview(data).cast('b')
        while buffer:
            try:
                nsent = self._socket_send(buffer, flags)
                buffer = buffer[nsent:]
            except WantWrite:
                await _write_wait(self._fileno)
            except WantRead:
                await _read_wait(self._fileno)

    async def accept(self):
        while True:
            try:
                client, addr = self._socket.accept()
                return Socket(client), addr
            except WantRead:
                await _read_wait(self._fileno)

    async def connect_ex(self, address):
        try:
            await self.connect(address)
            return 0
        except OSError as e:
            return e.errno

    async def connect(self, address):
        try:
            result = self._socket.connect(address)
            if getattr(self, 'do_handshake_on_connect', False):
                await self.do_handshake()
            return result
        except WantWrite:
            await _write_wait(self._fileno)
        err = self._socket.getsockopt(SOL_SOCKET, SO_ERROR)
        if err != 0:
            raise OSError(err, 'Connect call failed %s' % (address,))
        if getattr(self, 'do_handshake_on_connect', False):
            await self.do_handshake()

    async def recvfrom(self, buffersize, flags=0):
        while True:
            try:
                return self._socket.recvfrom(buffersize, flags)
            except WantRead:
                await _read_wait(self._fileno)
            except WantWrite:
                await _write_wait(self._fileno)

    async def recvfrom_into(self, buffer, bytes=0, flags=0):
        while True:
            try:
                return self._socket.recvfrom_into(buffer, bytes, flags)
            except WantRead:
                await _read_wait(self._fileno)
            except WantWrite:
                await _write_wait(self._fileno)

    async def sendto(self, bytes, flags_or_address, address=None):
        if address:
            flags = flags_or_address
        else:
            address = flags_or_address
            flags = 0
        while True:
            try:
                return self._socket.sendto(bytes, flags, address)
            except WantWrite:
                await _write_wait(self._fileno)
            except WantRead:
                await _read_wait(self._fileno)

    async def recvmsg(self, bufsize, ancbufsize=0, flags=0):
        while True:
            try:
                return self._socket.recvmsg(bufsize, ancbufsize, flags)
            except WantRead:
                await _read_wait(self._fileno)

    async def recvmsg_into(self, buffers, ancbufsize=0, flags=0):
        while True:
            try:
                return self._socket.recvmsg_into(buffers, ancbufsize, flags)
            except WantRead:
                await _read_wait(self._fileno)

    async def sendmsg(self, buffers, ancdata=(), flags=0, address=None):
        while True:
            try:
                return self._socket.sendmsg(buffers, ancdata, flags, address)
            except WantRead:
                await _write_wait(self._fileno)
    
    # Special functions for SSL
    async def do_handshake(self):
        while True:
            try:
                return self._socket.do_handshake()
            except WantRead:
                await _read_wait(self._fileno)
            except WantWrite:
                await _write_wait(self._fileno)

            
    # Design discussion.  Why make close() async?   Partly it's to make the
    # programming interface highly uniform with the other methods (all of which
    # involve an await).  It's also to provide consistency with the Stream
    # API below which requires an asynchronous close to properly flush I/O
    # buffers.

    async def close(self):
        if self._socket:
            self._socket.close()
        self._socket = None
        self._fileno = -1

    async def __aenter__(self):
        self._socket.__enter__()
        return self

    async def __aexit__(self, *args):
        if self._socket:
            self._socket.__exit__(*args)
      
    def __enter__(self):
        raise RuntimeError('Use async with')
    
    def __exit__(self, *args):
        pass

MAX_READ = 65536

class StreamBase(object):
    '''
    Base class for file-like objects.
    '''
    def __init__(self, fileobj):
        self._file = fileobj
        self._fileno = fileobj.fileno()
        self._buffer = bytearray()

    def __repr__(self):
        return '<curio.%s %r>' % (type(self).__name__, self._file)

    def fileno(self):
        return self._fileno

    # ---- Methods that must be implemented in child classes
    async def _read(self, maxbytes=-1):
        raise NotImplemented()

    async def write(self, data):
        raise NotImplemented()

    # ---- General I/O methods for streams
    async def read(self, maxbytes=-1):
        buf = self._buffer
        if buf:
            if maxbytes < 0 or len(buf) <= maxbytes:
                data = bytes(buf)
                buf.clear()
            else:
                data = bytes(buf[:n])
                del buf[:n]
        else:
            data = await self._read(maxbytes)
        return data

    async def readall(self):
        chunks = []
        if self._buffer:
            chunks.append(bytes(self._buffer))
            self._buffer.clear()
        while True:
            chunk = await self.read()
            if not chunk:
                return b''.join(chunks)
            chunks.append(chunk)

    async def read_exactly(self, nbytes):
        chunks = []
        while nbytes > 0:
            chunk = await self.read(nbytes)
            if not chunk:
                raise EOFError('Unexpected end of data')
            chunks.append(chunk)
            nbytes -= len(chunk)
        return b''.join(chunks)

    async def readline(self):
        while True:
            nl_index = self._buffer.find(b'\n')
            if nl_index >= 0:
                resp = bytes(self._buffer[:nl_index+1])
                del self._buffer[:nl_index+1]
                return resp
            data = await self._read(MAX_READ)
            if data == b'':
                resp = bytes(self._buffer)
                self._buffer.clear()
                return resp
            self._buffer.extend(data)

    async def readlines(self):
        lines = []
        async for line in self:
            lines.append(line)
        return lines

    async def writelines(self, lines):
        for line in lines:
            await self.write(line)

    async def flush(self):
         pass

    # Why async close()?   If the underlying file is buffered, the contents need
    # to be flushed first--a process that might cause a BlockingIOError.  In
    # that case, we have to suspend briefly until the buffers free up space.
    async def close(self):
        await self.flush()
        if self._file:
            self._file.close()
        self._file = None
        self._fileno = -1

    async def __aiter__(self):
        return self

    async def __anext__(self):
        line = await self.readline()
        if line:
            return line
        else:
            raise StopAsyncIteration

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        await self.close()

    def __iter__(self):
        raise RuntimeError('Use: async-for to iterate')

    def __enter__(self):
        raise RuntimeError('Use: async-with for context management')

    def __exit__(self, *args):
        pass

class FileStream(StreamBase):
    '''
    Wrapper around a file-like object.  File is put into non-blocking mode.
    The underlying file must be in binary mode.  
    '''
    def __init__(self, fileobj):
        assert not isinstance(fileobj, io.TextIOBase), 'Only binary mode files allowed'
        super().__init__(fileobj)
        os.set_blocking(self._fileno, False)

        # Common bound methods
        self._file_read = fileobj.read
        self._file_write = fileobj.write

    @contextmanager
    def blocking(self):
        '''
        Allow temporary access to the underlying file in blocking mode
        '''
        if self._buffer:
            raise IOError('There is unread buffered data.')
        try:
            os.set_blocking(self._fileno, True)
            yield self._file
        finally:
            os.set_blocking(self._fileno, False)

    async def _read(self, maxbytes=-1):
        while True:
            # In non-blocking mode, a file-like object might return None if no data is
            # available.  Alternatively, we'll catch the usual blocking exceptions just to be safe
            try:
                data = self._file_read(maxbytes)
                if data is None:
                    await _read_wait(self._fileno)
                else:
                    return data
            except WantRead:
                await _read_wait(self._fileno)
            except WantWrite:
                await _write_wait(self._fileno)

    async def write(self, data):
        nwritten = 0
        view = memoryview(data).cast('b')
        while view:
            try:
                nbytes = self._file_write(view)
                if nbytes is None:
                    raise BlockingIOError()
                nwritten += nbytes
                view = view[nbytes:]
            except WantWrite as e:
                if hasattr(e, 'characters_written'):
                    nwritten += e.characters_written
                    view = view[e.characters_written:]
                await _write_wait(self._fileno)
            except WantRead:
                await _read_wait(self._fileno)

        return nwritten

    async def flush(self):
        while True:
            try:
                return self._file.flush() 
            except WantWrite:
                await _write_wait(self._fileno)
            except WantRead:
                await _read_wait(self._fileno)

class SocketStream(StreamBase):
    '''
    Stream wrapper for a socket.
    '''
    def __init__(self, sock):
        super().__init__(sock)
        sock.setblocking(False)

        # Common bound methods
        self._socket_recv = sock.recv
        self._socket_send = sock.send

    @contextmanager
    def blocking(self):
        '''
        Allow temporary access to the underlying file in blocking mode
        '''
        if self._buffer:
            raise IOError('There is unread buffered data.')
        try:
            self._file.setblocking(True)
            yield open(self._fileno, 'rb+', buffering=0, closefd=False)
        finally:
            self._file.setblocking(False)

    async def _read(self, maxbytes=-1):
        while True:
            try:
                data = self._socket_recv(maxbytes if maxbytes > 0 else MAX_READ)
                return data
            except WantRead:
                await _read_wait(self._fileno)
            except WantWrite:
                await _write_wait(self._fileno)

    async def write(self, data):
        nwritten = 0
        view = memoryview(data).cast('b')
        while view:
            try:
                nbytes = self._socket_send(view)
                nwritten += nbytes
                view = view[nbytes:]
            except WantWrite:
                await _write_wait(self._fileno)
            except WantRead:
                await _read_wait(self._fileno)
        return nwritten
    



# curio/task.py
#
# Task class and task related functions.

__all__ = [ 'sleep', 'current_task', 'spawn', 'gather', 'timeout_after', 'ignore_after' ]

from .errors import CancelledError, TaskTimeout, TaskError
from .traps import *

class Task(object):
    '''
    The Task class wraps a coroutine and provides some additional attributes
    related to execution state and debugging.
    '''
    __slots__ = (
        'id', 'daemon',  'coro', '_send', '_throw', 'cycles', 'state',
        'cancel_func', 'future', 'sleep', 'timeout', 'exc_info', 'next_value',
        'next_exc', 'joining', 'cancelled', 'terminated', '_last_io', '__weakref__',
        )
    _lastid = 1
    def __init__(self, coro, daemon=False, taskid=None):
        if taskid is None:
            taskid = Task._lastid   
            Task._lastid += 1
        self.id = taskid
        self.coro = coro           # Underlying generator/coroutine
        self.daemon = daemon       # Daemonic flag
        self.cycles = 0            # Execution cycles completed
        self.state = 'INITIAL'     # Execution state
        self.cancel_func = None    # Cancellation function
        self.future = None         # Pending Future (if any)
        self.sleep = None          # Pending sleep (if any)
        self.timeout = None        # Pending timeout (if any)
        self.exc_info = None       # Exception info (if any on crash)
        self.next_value = None     # Next value to send on execution
        self.next_exc = None       # Next exception to send on execution
        self.joining = None        # Optional set of tasks waiting to join with this one
        self.cancelled = False     # Cancelled?
        self.terminated = False    # Terminated?
        self._last_io = None       # Last I/O operation performed
        self._send = coro.send     # Bound coroutine methods
        self._throw = coro.throw

    def __repr__(self):
        return 'Task(id=%r, %r, state=%r)' % (self.id, self.coro, self.state)

    def __str__(self):
        return self.coro.__qualname__

    def __del__(self):
        self.coro.close()

    async def join(self):
        '''
        Wait for a task to terminate.  Returns the return value (if any)
        or raises a TaskError if the task crashed with an exception.
        '''
        await _join_task(self)
        if self.exc_info:
            raise TaskError('Task crash') from self.exc_info[1]
        else:
            return self.next_value

    async def cancel(self):
        '''
        Cancel a task by raising a CancelledError exception.  Does not
        return until the task actually terminates.  Returns True if
        the task was actually cancelled. False is returned if the task
        was already completed.  
        '''
        if self.terminated:
            return False
        else:
            await _cancel_task(self)
            return True

# ----------------------------------------------------------------------
# Public-facing task-related functions.  Some of these functions are
# merely a layer over a low-level trap using async/await.  One reason
# for doing this is that the user will get a more proper warning message
# if they use the function without using the required 'await' keyword.
# -----------------------------------------------------------------------

async def current_task():
    '''
    Returns a reference to the current task
    '''
    return await _get_current()

async def sleep(seconds):
    '''
    Sleep for a specified number of seconds.  Sleeping for 0 seconds
    makes a task immediately switch to the next ready task (if any).
    '''
    await _sleep(seconds)

async def spawn(coro, *, daemon=False):
    '''
    Create a new task.  Use the daemon=True option if the task runs
    forever as a background task.
    '''
    return await _spawn(coro, daemon)

async def gather(tasks, *, return_exceptions=False):
    '''
    Wait for and gather results from a collection of tasks.
    '''
    results = []
    for task in tasks:
        try:
            results.append(await task.join())
        except Exception as e:
            if return_exceptions:
                results.append(e)
            else:
                raise
    return results

# Helper class for running timeouts as a context manager
class _TimeoutAfter(object):
    def __init__(self, seconds, ignore=False, timeout_result=None):
        self._seconds = seconds
        self._ignore = ignore
        self._timeout_result = timeout_result
        self.result = True

    async def __aenter__(self):
        self._prior = await _set_timeout(self._seconds)
        return self

    async def __aexit__(self, ty, val, tb):
        await _unset_timeout(self._prior)
        if ty == TaskTimeout:
            self.result = self._timeout_result
            if self._ignore:
                return True

async def _timeout_after_func(seconds, coro, ignore=False, timeout_result=None):
    prior = await _set_timeout(seconds)
    try:
        return await coro
    except TaskTimeout:
        if not ignore:
            raise
        return timeout_result
    finally:
        await _unset_timeout(prior)

def timeout_after(seconds, coro=None):
    '''
    Raise a TaskTimeout exception in the calling task after seconds
    have elapsed.  This function may be used in two ways. You can
    apply it to the execution of a single coroutine:

         await timeout_after(seconds, coro(args))

    or you can use it as an asynchronous context manager to apply
    a timeout to a block of statements:

         async with timeout_after(seconds):
             await coro1(args)
             await coro2(args)
             ...
    '''
    if coro is None:
        return _TimeoutAfter(seconds)
    else:
        return _timeout_after_func(seconds, coro)

def ignore_after(seconds, coro=None, *, timeout_result=None):
    '''
    Stop the enclosed task or block of code after seconds have
    elapsed.  No exception is raised when time expires. Instead, None
    is returned.  This is often more convenient that catching an
    exception.  You can apply the function to a single coroutine:
    
        if ignore_after(5, coro(args)) is None:
            # A timeout occurred
            ...

    Alternatively, you can use this function as an async context
    manager on a block of statements like this:

        async with ignore_after(5) as r:
            await coro1(args)
            await coro2(args)
            ...
        if r.result is None:
            # A timeout occurred

    When used as a context manager, the return manager object has
    a result attribute that will be set to None if the time
    period expires (or True otherwise). 

    You can change the return result to a different value using
    the timeout_result keyword argument.
    '''
    if coro is None:
        return _TimeoutAfter(seconds, ignore=True, timeout_result=timeout_result)
    else:
        return _timeout_after_func(seconds, coro, ignore=True, timeout_result=timeout_result)



# udp_echo
#
# An example of a UDP echo client

import curio

async def main(addr):
    sock = await curio.open_unix_connection(addr)
    for i in range(1000):
        msg = ('Message %d' % i).encode('ascii')
        print(msg)
        await sock.send(msg)
        resp = await sock.recv(1000)
        assert msg == resp
    await sock.close()

if __name__ == '__main__':
    curio.run(main('/tmp/curiounixecho'))



# Example of making an SSL connection and downloading data

import curio

async def main():
    sock = await curio.open_connection('www.python.org', 443,
                                       ssl=True, server_hostname='www.python.org')
    async with sock:
        await sock.sendall(b'GET / HTTP/1.0\r\nHost: www.python.org\r\n\r\n')
        chunks = []
        while True:
            chunk = await sock.recv(10000)
            if not chunk:
                break
            chunks.append(chunk)
        
    response = b''.join(chunks)
    print(response.decode('latin-1'))


if __name__ == '__main__':
    curio.run(main())



# curio/network.py
#
# Some high-level functions useful for writing network code.  These are loosely
# based on their similar counterparts in the asyncio library. Some of the
# fiddly low-level bits are borrowed.   

__all__ = [ 
    'open_connection',
    'tcp_server',
    'open_unix_connection',
    'unix_server'
    ]

from . import socket
from . import ssl as curiossl
from .task import spawn

async def _wrap_ssl_client(sock, ssl, server_hostname):
    # Applies SSL to a client connection. Returns an SSL socket.
    if ssl:
        if isinstance(ssl, bool):
            sslcontext = curiossl.create_default_context()
            if not server_hostname:
                sslcontext._context.check_hostname = False
                sslcontext._context.verify_mode = curiossl.CERT_NONE
        else:
            # Assume that ssl is an already created context
            sslcontext = ssl

        if server_hostname:
            extra_args = {'server_hostname': server_hostname}
        else:
            extra_args = { }

        sock = sslcontext.wrap_socket(sock, **extra_args)
        await sock.do_handshake()
    return sock

async def open_connection(host, port, *, ssl=None, source_addr=None, server_hostname=None):
    '''
    Create a TCP connection to a given Internet host and port with optional SSL applied to it.
    '''
    if server_hostname and not ssl:
        raise ValueError('server_hostname is only applicable with SSL')

    sock = await socket.create_connection((host,port), source_addr)

    try:
        # Apply SSL wrapping to the connection, if applicable
        if ssl:
            sock = await _wrap_ssl_client(sock, ssl, server_hostname)

        return sock
    except Exception:
        sock._socket.close()
        raise

async def open_unix_connection(path, *, ssl=None, server_hostname=None):
    if server_hostname and not ssl:
        raise ValueError('server_hostname is only applicable with SSL')

    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    try:
        await sock.connect(path)

        # Apply SSL wrapping to connection, if applicable
        if ssl:
            sock = await _wrap_ssl_client(sock, ssl, server_hostname)

        return sock
    except Exception:
        sock._socket.close()
        raise

class Server(object):
    def __init__(self, sock, client_connected_task, ssl=None):
        self._sock = sock
        self._ssl = ssl
        self._client_connected_task = client_connected_task
        self._task = None

    async def serve_forever(self):
        async with self._sock:
            while True:
                client, addr = await self._sock.accept()

                if self._ssl:
                    client = self._ssl.wrap_socket(client, server_side=True, do_handshake_on_connect=False)

                client_task = await spawn(self.run_client(client, addr))
                del client

    async def run_client(self, client, addr):
        async with client:
            await self._client_connected_task(client, addr)

    async def cancel(self):
        if self._task:
            await self._task.cancel()


async def _run_server(sock, client_connected_task, ssl=None):
    async def run_client(client, addr):
        async with client:
              await client_connected_task(client, addr)

    async with sock:
        while True:
            client, addr = await sock.accept()
            if ssl:
                client = ssl.wrap_socket(client, server_side=True, do_handshake_on_connect=False)
            client_task = await spawn(run_client(client, addr))
            del client
        
async def tcp_server(host, port, client_connected_task, *, 
                     family=socket.AF_INET, backlog=100, ssl=None, reuse_address=True):

    if ssl and not isinstance(ssl, curiossl.CurioSSLContext):
        raise ValueError('ssl argument must be a curio.ssl.SSLContext instance')

    sock = socket.socket(family, socket.SOCK_STREAM)
    try:
        if reuse_address:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)

        sock.bind((host, port))
        sock.listen(backlog)
        await _run_server(sock, client_connected_task, ssl)
    except Exception:
        sock._socket.close()
        raise
    
async def unix_server(path, client_connected_task, *, backlog=100, ssl=None):
    if ssl and not isinstance(ssl, curiossl.CurioSSLContext):
        raise ValueError('ssl argument must be a curio.ssl.SSLContext instance')
  
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    try:
        sock.bind(path)
        sock.listen(backlog)
        await _run_server(sock, client_connected_task, ssl)
    except Exception:
        sock._socket.close()
        raise




from tornado.ioloop import IOLoop
from tornado.tcpserver import TCPServer

class EchoServer(TCPServer):
    def handle_stream(self, stream, address):
        self._stream = stream
        self._stream.read_until_close(None, self.handle_read)

    def handle_read(self, data):
        self._stream.write(data)

if __name__ == '__main__':
    server = EchoServer()
    server.bind(25000)
    server.start(1)
    IOLoop.instance().start()
    IOLoop.instance().close()



# A simple echo server with threads

from socket import *
from threading import Thread

def echo_server(addr):
    sock = socket(AF_INET, SOCK_STREAM)
    sock.setsockopt(SOL_SOCKET, SO_REUSEADDR,1)
    sock.bind(addr)
    sock.listen(5)
    while True:
        client, addr = sock.accept()
        Thread(target=echo_handler, args=(client, addr), daemon=True).start()

def echo_handler(client, addr):
    print('Connection from', addr)
    client.setsockopt(IPPROTO_TCP, TCP_NODELAY, 1)
    with client:
        while True:
            data = client.recv(100000)
            if not data:
                break
            client.sendall(data)
    print('Connection closed')

if __name__ == '__main__':
    echo_server(('',25000))



# Example: A simple echo server written directly with asyncio

import asyncio
from socket import *

async def echo_server(loop, address):
    sock = socket(AF_INET, SOCK_STREAM)
    sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
    sock.bind(address)
    sock.listen(5)
    sock.setblocking(False)
    print('Server listening at', address)
    with sock:
        while True:
             client, addr = await loop.sock_accept(sock)
             print('Connection from', addr)
             loop.create_task(echo_client(loop, client))

async def echo_client(loop, client):
    client.setsockopt(IPPROTO_TCP, TCP_NODELAY, 1)
    with client:
         while True:
             data = await loop.sock_recv(client, 10000)
             if not data:
                  break
             await loop.sock_sendall(client, data)
    print('Connection closed')

if __name__ == '__main__':
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    loop.set_debug(False)
    loop.create_task(echo_server(loop, ('',25000)))
    loop.run_forever()



# test_subprocess.py

import sys
from curio import subprocess
from curio import *
import os

# ---- Test subprocesses and worker task related functions

executable = sys.executable
dirname = os.path.dirname(__file__)

def test_simple(kernel):
    results = []
    async def subproc():
#        out = await subprocess.run([executable, '-m', 'curio.test.slow'], stdout=subprocess.PIPE)
        out = await subprocess.run([executable, os.path.join(dirname, 'child.py')], stdout=subprocess.PIPE)
        results.append(out.stdout)
        results.append(out.returncode)

    kernel.run(subproc())
    assert results == [
            b't-minus 4\nt-minus 3\nt-minus 2\nt-minus 1\n',
            0,
            ]

def test_simple_check_output(kernel):
    results = []
    async def subproc():
#        out = await subprocess.check_output([executable, '-m', 'curio.test.slow'])
        out = await subprocess.check_output([executable, os.path.join(dirname, 'child.py')])
        results.append(out)

    kernel.run(subproc())
    assert results == [
            b't-minus 4\nt-minus 3\nt-minus 2\nt-minus 1\n',
            ]

def test_bad_cmd(kernel):
    results = []
    async def subproc():
        try:
            out = await subprocess.run([executable, '-m', 'curio.test.bad'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
            results.append('what?')
        except subprocess.CalledProcessError:
            results.append('bad command')

    kernel.run(subproc())
    assert results == ['bad command']

def test_bad_cmd_check_output(kernel):
    results = []
    async def subproc():
        try:
            out = await subprocess.check_output([executable, '-m', 'curio.test.bad'], stderr=subprocess.STDOUT)
            results.append('what?')
        except subprocess.CalledProcessError:
            results.append('bad command')

    kernel.run(subproc())
    assert results == [ 'bad command' ]

def test_timeout(kernel):
    results = []
    async def subproc():
        try:
#            out = await subprocess.run([executable, '-m', 'curio.test.slow'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=1)
            out = await subprocess.run([executable, os.path.join(dirname, 'child.py')], stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=1)
            results.append('what?')
        except subprocess.TimeoutExpired:
            results.append('timeout')

    kernel.run(subproc())
    assert results == [ 'timeout' ]



# process.py
#
# Performance test of submitting work to a process pool. 

import time
import curio
from multiprocessing import Pool
from concurrent.futures import ProcessPoolExecutor
import asyncio

COUNT = 10000

def fib(n):
    if n <= 2:
        return 1
    else:
        return fib(n-1) + fib(n-2)

def curio_test(x):
    async def main():
        for n in range(COUNT):
            r = await curio.run_in_process(fib, x)
    start = time.time()
    curio.run(main())
    end = time.time()
    print('Curio:', end-start)

def mp_test(x):
    pool = Pool()
    def main():
        for n in range(COUNT):
            r = pool.apply(fib, (x,))
    start = time.time()
    main()
    end = time.time()
    print('multiprocessing:', end-start)

def future_test(x):
    pool = ProcessPoolExecutor()
    def main():
        for n in range(COUNT):
            f = pool.submit(fib, x)
            r = f.result()
    start = time.time()
    main()
    end =  time.time()
    print('concurrent.futures:', end-start)

def asyncio_test(x):
    pool = ProcessPoolExecutor()
    async def main(loop):
        for n in range(COUNT):
            r = await loop.run_in_executor(pool, fib, x)

    loop = asyncio.get_event_loop()
    start = time.time()
    loop.run_until_complete(asyncio.ensure_future(main(loop)))
    end = time.time()
    print('asyncio:', end-start)

def uvloop_test(x):
    try:
        import uvloop
    except ImportError:
        return

    pool = ProcessPoolExecutor()
    async def main(loop):
        for n in range(COUNT):
            r = await loop.run_in_executor(pool, fib, x)

    loop = uvloop.new_event_loop()
    asyncio.set_event_loop(loop)
    start = time.time()
    loop.run_until_complete(asyncio.ensure_future(main(loop)))
    end = time.time()
    print('uvloop:', end-start)
    
if __name__ == '__main__':
    import sys
    if len(sys.argv) != 2:
        raise SystemExit('Usage: %s n' % sys.argv[0])
    x = int(sys.argv[1])
    asyncio_test(x)
    uvloop_test(x)
    future_test(x)
    mp_test(x)
    curio_test(x)




        



# curio/sync.py
#
# Implementation of common task synchronization primitives such as
# events, locks, semaphores, and condition variables. These primitives
# are only safe to use in the curio framework--they are not thread safe.

__all__ = ['Event', 'Lock', 'Semaphore', 'BoundedSemaphore', 'Condition', 'abide' ]

import threading
from inspect import iscoroutinefunction

from .traps import _wait_on_queue, _reschedule_tasks
from .kernel import kqueue
from . import workers
from .errors import CancelledError, TaskTimeout
from .task import spawn, timeout_after

class Event(object):
    __slots__ = ('_set', '_waiting')
    def __init__(self):
        self._set = False
        self._waiting = kqueue()

    def __repr__(self):
        res = super().__repr__()
        extra = 'set' if self._set else 'unset'
        return '<{} [{},waiters:{}]>'.format(res[1:-1], extra, len(self._waiting))

    def is_set(self):
        return self._set

    def clear(self):
        self._set = False

    async def wait(self):
        if self._set:
            return
        await _wait_on_queue(self._waiting, 'EVENT_WAIT')

    async def set(self):
        self._set = True
        await _reschedule_tasks(self._waiting, len(self._waiting))

class _LockBase(object):
    async def __aenter__(self):
        await self.acquire()
        return None

    async def __aexit__(self, exc_type, exc, tb):
        await self.release()

    def __enter__(self):
        raise RuntimeError('Use async with')

    def __exit__(self, *args):
        pass

class Lock(_LockBase):
    __slots__ = ('_acquired', '_waiting')
    def __init__(self):
        self._acquired = False
        self._waiting = kqueue()

    def __repr__(self):
        res = super().__repr__()
        extra = 'locked' if self.locked() else 'unlocked'
        return '<{} [{},waiters:{}]>'.format(res[1:-1], extra, len(self._waiting))

    async def acquire(self):
        if self._acquired:
            await _wait_on_queue(self._waiting, 'LOCK_ACQUIRE')
        self._acquired = True
        return True

    async def release(self):
        assert self._acquired, 'Lock not acquired'
        if self._waiting:
            await _reschedule_tasks(self._waiting, n=1)
        else:
            self._acquired = False

    def locked(self):
        return self._acquired

class Semaphore(_LockBase):
    __slots__ = ('_value', '_waiting')
    def __init__(self, value=1):
        self._value = value
        self._waiting = kqueue()

    def __repr__(self):
        res = super().__repr__()
        extra = 'locked' if self.locked() else 'unlocked'
        return '<{} [{},value:{},waiters:{}]>'.format(res[1:-1], extra, self._value, len(self._waiting))

    async def acquire(self):
        if self._value <= 0:
            await _wait_on_queue(self._waiting, 'SEMA_ACQUIRE')
        else:
            self._value -= 1
        return True

    async def release(self):
        if self._waiting:
            await _reschedule_tasks(self._waiting, n=1)
        else:
            self._value += 1
        
    def locked(self):
        return self._value == 0

class BoundedSemaphore(Semaphore):
    __slots__ = ('_bound_value',)
    def __init__(self, value=1):
        self._bound_value = value
        super().__init__(value)

    async def release(self):
        if self._value >= self._bound_value:
            raise ValueError('BoundedSemaphore released too many times')
        await super().release()

class Condition(_LockBase):
    __slots__ = ('_lock', '_waiting')
    def __init__(self, lock=None):
        if lock is None:
            self._lock = Lock()
        else:
            self._lock = lock
        self._waiting = kqueue()

    def __repr__(self):
        res = super().__repr__()
        extra = 'locked' if self.locked() else 'unlocked'
        return '<{} [{},waiters:{}]>'.format(res[1:-1], extra, len(self._waiting))

    def locked(self):
        return self._lock.locked()

    async def acquire(self):
        await self._lock.acquire()

    async def release(self):
        await self._lock.release()

    async def wait(self):
        if not self.locked():
            raise RuntimeError("Can't wait on unacquired lock")
        await self.release()
        try:
            await _wait_on_queue(self._waiting, 'COND_WAIT')
        finally:
            await self.acquire()

    async def wait_for(self, predicate):
        while True:
            result = predicate()
            if result:
                return result
            await self.wait()

    async def notify(self, n=1):
        if not self.locked():
            raise RuntimeError("Can't notify on unacquired lock")
        await _reschedule_tasks(self._waiting, n=n)

    async def notify_all(self):
        await self.notify(len(self._waiting))

# Class that adapts a synchronous context-manager to an asynchronous manager
class _contextadapt(object):
    def __init__(self, manager):
        self.manager = manager

    async def __aenter__(self):
        task = await spawn(workers.run_in_thread(self.manager.__enter__))
        try:
            return await task.join()
        except (CancelledError, TaskTimeout) as err:
            # An interesting corner case... if we're cancelled why waiting to
            # enter, we'd better arrange to exit in case it eventually succeeds.
            async def manager_release(task, err):
                  await task.join()
                  await workers.run_in_thread(self.manager.__exit__, type(err), err, err.__traceback__)

            await spawn(manager_release(task, err))
            raise

    async def __aexit__(self, *args):
        return await workers.run_in_thread(self.manager.__exit__, *args)

def abide(op, *args, **kwargs):
    '''
    Make curio abide by the execution requirements of a given
    function, coroutine, or context manager.  If op is coroutine
    function, it is called with the given arguments.  If op is an
    asynchronous context manager, it is returned unmodified.  If op is
    a synchronous function, it is executed in a separate thread.  If
    op is a synchronous context manager, it is wrapped by an
    asynchronous context manager that executes the __enter__() and
    __exit__() methods in threads.

    The main use of this function is in code that wants to safely
    synchronize curio with threads and processes. For example, if you
    write code this like:

        async with abide(lck):
            statements

    The code will work correctly if lck is an async lock defined by curio or
    a foreign lock defined by the threading or multiprocessing modules.

    You can also use abide() with method calls. For example:

        await abide(q.put, item)

    would safely execute a put(item) method on a queue regardless of
    whether or not q is a curio queue or a queue used for threads.
    '''
    if iscoroutinefunction(op):
        return op(*args, **kwargs)

    if hasattr(op, '__aexit__'):
        return op

    if hasattr(op, '__exit__'):
        return _contextadapt(op)

    if not callable(op):
        raise TypeError('%r object is not callable' % type(op).__name__)

    return workers.run_in_thread(op, *args, **kwargs)



# curio/signal.py
#
# Signal sets and signal related functionality

__all__ = [ 'SignalSet' ]

from contextlib import contextmanager
from collections import deque
import signal

from .traps import *

class SignalSet(object):
    def __init__(self, *signos):
        self.signos = signos          # List of all signal numbers being tracked
        self.pending = deque()        # Pending signals received
        self.waiting = None           # Task waiting for the signals (if any)
        self.watching = False         # Are the signals being watched right now?
    
    async def __aenter__(self):
        assert not self.watching
        await _sigwatch(self)
        self.watching = True
        return self

    async def __aexit__(self, *args):
        await _sigunwatch(self)
        self.watching = False

    def __enter__(self):
        raise RuntimeError('Use async with')

    def __exit__(self, *args):
        pass

    async def wait(self):
        '''
        Wait for a single signal from the signal set to arrive.
        '''
        if not self.watching:
            async with self:
                return await self.wait()

        while True:
            if self.pending:
                return self.pending.popleft()
            await _sigwait(self)

    @contextmanager
    def ignore(self):
        '''
        Context manager. Temporarily ignores all signals in the signal set. 
        '''
        try:
            orig_signals = [ (signo, signal.signal(signo, signal.SIG_IGN)) for signo in self.signos ]
            yield
        finally:
            for signo, handler in orig_signals:
                signal.signal(signo, handler)



# Example: A simple echo server written directly with sockets

from curio import run, spawn
from curio.socket import *

async def echo_server(address):
    sock = socket(AF_INET, SOCK_STREAM)
    sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
    sock.bind(address)
    sock.listen(5)
    print('Server listening at', address)
    async with sock:
        while True:
             client, addr = await sock.accept()
             print('Connection from', addr)
             await spawn(echo_client(client))

async def echo_client(client):
    async with client:
         while True:
             data = await client.recv(10000)
             if not data:
                  break
             await client.sendall(data)
    print('Connection closed')

if __name__ == '__main__':
     run(echo_server(('',25000)))



# channel.py
#
# Support for a message passing channel that can send bytes or pickled
# Python objects on a stream.  Compatible with the Connection class in the
# multiprocessing module, but rewritten for a purely asynchronous runtime.

__all__ = ['Channel', 'Listener', 'Client']

import os
import pickle
import struct
import io
import hmac
import multiprocessing.connection as mpc
from contextlib import contextmanager

from . import socket
from .errors import CurioError
from .io import StreamBase, FileStream

# Authentication parameters (copied from multiprocessing)

import multiprocessing.connection as mpc

AUTH_MESSAGE_LENGTH = mpc.MESSAGE_LENGTH    # 20
CHALLENGE = mpc.CHALLENGE                   # b'#CHALLENGE#'
WELCOME = mpc.WELCOME                       # b'#WELCOME#'
FAILURE = mpc.FAILURE                       # b'#FAILURE#'

class ChannelError(CurioError):
    pass

class AuthenticationError(ChannelError):
    pass

class Channel(object):
    '''
    A communication channel for sending size-prefixed messages of bytes
    or pickled Python objects.  Must be passed a pair of reader/writer
    streams for performing the underlying communication.
    '''
    def __init__(self, reader, writer):
        assert isinstance(reader, StreamBase) and isinstance(writer, StreamBase)
        self._reader = reader
        self._writer = writer

    @classmethod
    def from_Connection(cls, conn):
        '''
        Creates a channel from a multiprocessing Connection. Note: The
        multiprocessing connection is detached by having its handle set to None.
        
        This method can be used to make curio talk over Pipes as created by
        multiprocessing.  For example:

              p1, p2 = multiprocessing.Pipe()
              p1 = Channel.from_Connection(p1)
              p2 = Channel.from_Connection(p2)

        '''
        assert isinstance(conn, mpc._ConnectionBase)
        reader = FileStream(open(conn._handle, 'rb', buffering=0))
        writer = FileStream(open(conn._handle, 'wb', buffering=0, closefd=False))
        conn._handle = None
        return cls(reader, writer)

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        await self.close()
    
    async def close(self):
        await self._reader.close()
        await self._writer.close()

    async def send_bytes(self, buf, offset=0, size=None):
        '''
        Send a buffer of bytes as a single message
        '''
        m = memoryview(buf)
        if m.itemsize > 1:
            m = memoryview(bytes(m))
        n = len(m)
        if offset < 0:
            raise ValueError("offset is negative")
        if n < offset:
            raise ValueError("buffer length < offset")
        if size is None:
            size = n - offset
        elif size < 0:
            raise ValueError("size is negative")
        elif offset + size > n:
            raise ValueError("buffer length < offset + size")

        header = struct.pack('!i', size)
        if size >= 16384:
            await self._writer.write(header)
            await self._writer.write(m[offset:offset+size])
        else:
            msg = header + bytes(m[offset:offset+size])
            await self._writer.write(msg)

    async def recv_bytes(self, maxlength=None):
        '''
        Receive a message of bytes as a single message.
        '''
        header = await self._reader.read_exactly(4)
        size,  = struct.unpack('!i', header)
        if maxlength is not None:
            if size > maxlength:
                raise IOError('Message too large. %d bytes > %d maxlength' % (size, maxlength))

        msg = await self._reader.read_exactly(size)
        return msg

    async def recv_bytes_into(self, buf, offset=0):
        '''
        Receive bytes into a writable memory buffer
        '''
        pass

    async def send(self, obj):
        '''
        Send an arbitrary Python object. Uses pickle to serialize.
        '''
        await self.send_bytes(pickle.dumps(obj))

    async def recv(self):
        '''
        Receive a Python object. Uses pickle to unserialize.
        '''
        msg = await self.recv_bytes()
        return pickle.loads(msg)

    async def _deliver_challenge(self, authkey):
        message = os.urandom(AUTH_MESSAGE_LENGTH)
        await self.send_bytes(CHALLENGE + message)
        digest = hmac.new(authkey, message, 'md5').digest()
        response = await self.recv_bytes(maxlength=256)
        if response == digest:
            await self.send_bytes(WELCOME)
        else:
            await self.send_bytes(FAILURE)
            raise AuthenticationError('digest received was wrong')

    async def _answer_challenge(self, authkey):
        message = await self.recv_bytes(maxlength=256)
        assert message[:len(CHALLENGE)] == CHALLENGE, 'message = %r' % message
        message = message[len(CHALLENGE):]
        digest = hmac.new(authkey, message, 'md5').digest()
        await self.send_bytes(digest)
        response = await self.recv_bytes(maxlength=256)

        if response != WELCOME:
            raise AuthenticationError('digest sent was rejected')

    async def authenticate_server(self, authkey):
        await self._deliver_challenge(authkey)
        await self._answer_challenge(authkey)

    async def authenticate_client(self, authkey):
        await self._answer_challenge(authkey)
        await self._deliver_challenge(authkey)

class Listener(object):
    def __init__(self, address, family=socket.AF_INET, backlog=1, authkey=None):
        self._sock = socket.socket(family, socket.SOCK_STREAM)
        if family == socket.AF_INET:
            self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
        self._sock.bind(address)
        self._sock.listen(backlog)
        self._authkey = authkey

    async def accept(self):
        client, addr = await self._sock.accept()
        fileno = client.detach()
        ch = Channel(FileStream(open(fileno, 'rb', buffering=0)),
                     FileStream(open(fileno, 'wb', buffering=0, closefd=False)))
        if self._authkey:
            await ch.authenticate_server(self._authkey)
        return ch

    async def close(self):
        await self._sock.close()

async def Client(address, family=socket.AF_INET, authkey=None):
    sock = socket.socket(family, socket.SOCK_STREAM)
    await sock.connect(address)
    fileno = sock.detach()
    ch = Channel(FileStream(open(fileno, 'rb', buffering=0)),
                 FileStream(open(fileno, 'wb', buffering=0, closefd=False)))
    if authkey:
        await ch.authenticate_client(authkey)
    return ch



# test_signal.py

import time
from curio import *
import signal, os

def test_task_signal(kernel):
    results = []

    async def child():
        async with SignalSet(signal.SIGUSR1, signal.SIGUSR2) as sig:
            signo = await sig.wait()
            results.append(signo)
            signo = await sig.wait()
            results.append(signo)

    async def main():
        task = await spawn(child())
        await sleep(0.1)
        results.append('sending USR1')
        os.kill(os.getpid(), signal.SIGUSR1)
        await sleep(0.1)
        results.append('sending USR2')
        os.kill(os.getpid(), signal.SIGUSR2)
        await sleep(0.1)
        results.append('done')

    kernel.run(main())
    assert results == [
            'sending USR1',
            signal.SIGUSR1,
            'sending USR2',
            signal.SIGUSR2,
            'done',
            ]

def test_task_signal_waitone(kernel):
    results = []

    async def child():
        sig = SignalSet(signal.SIGUSR1, signal.SIGUSR2)
        signo = await sig.wait()
        results.append(signo)
        signo = await sig.wait()
        results.append(signo)

    async def main():
        task = await spawn(child())
        await sleep(0.1)
        results.append('sending USR1')
        os.kill(os.getpid(), signal.SIGUSR1)
        await sleep(0.1)
        results.append('sending USR2')
        os.kill(os.getpid(), signal.SIGUSR2)
        await sleep(0.1)
        results.append('done')

    kernel.run(main())
    assert results == [
            'sending USR1',
            signal.SIGUSR1,
            'sending USR2',
            signal.SIGUSR2,
            'done',
            ]

def test_task_signal_ignore(kernel):
    results = []

    async def child():
        sig = SignalSet(signal.SIGUSR1, signal.SIGUSR2)
        async with sig:
             signo = await sig.wait()
             results.append(signo)
             with sig.ignore():
                 await sleep(1)
             results.append('here')

    async def main():
        task = await spawn(child())
        await sleep(0.1)
        results.append('sending USR1')
        os.kill(os.getpid(), signal.SIGUSR1)
        await sleep(0.5)
        results.append('sending USR1')
        os.kill(os.getpid(), signal.SIGUSR1)
        await sleep(0.1)
        await task.join()
        results.append('done')

    kernel.run(main())
    assert results == [
            'sending USR1',
            signal.SIGUSR1,
            'sending USR1',
            'here',
            'done',
            ]

def test_task_signal_timeout(kernel):
    results = []

    async def child():
        async with SignalSet(signal.SIGUSR1, signal.SIGUSR2) as sig:
            try:
                signo = await timeout_after(1.0, sig.wait())
                results.append(signo)
            except TaskTimeout:
                results.append('timeout')

    async def main():
        task = await spawn(child())
        await task.join()
        results.append('done')

    kernel.run(main())
    assert results == [
            'timeout',
            'done',
            ]



# test_io.py

from curio import *
from curio.socket import *
import io
import socket as std_socket

def test_socket_blocking(kernel):
    '''
    Test of exposing a socket in blocking mode
    '''
    results = []

    def sync_client(sock):
        assert isinstance(sock, std_socket.socket)
        data = sock.recv(8192)
        results.append(('client', data))
        sock.sendall(data)

    async def handler(client, addr):
        results.append('handler start')
        await client.send(b'OK')
        with client.blocking() as s:
            await run_in_thread(sync_client, s)
        results.append('handler done')

    async def test_client(address, serv):
        sock = socket(AF_INET, SOCK_STREAM)
        await sock.connect(address)
        await sock.recv(8)
        await sock.send(b'Message')
        data = await sock.recv(8192)
        await sock.close()
        await serv.cancel()

    async def main():
       serv = await spawn(tcp_server('',25000, handler))
       await spawn(test_client(('localhost', 25000), serv))

    kernel.run(main())

    assert results == [
            'handler start',
            ('client', b'Message'),
            'handler done'
            ]

def test_socketstream_blocking(kernel):
    '''
    Test of exposing a socket stream in blocking mode
    '''
    results = []
    
    def sync_client(f):
        assert isinstance(f, io.RawIOBase)
        data = f.read(8192)
        results.append(('client', data))
        f.write(data)

    async def handler(client, addr):
        results.append('handler start')
        await client.send(b'OK')
        s = client.as_stream()
        with s.blocking() as f:
            await run_in_thread(sync_client, f)
        results.append('handler done')

    async def test_client(address, serv):
        sock = socket(AF_INET, SOCK_STREAM)
        await sock.connect(address)
        await sock.recv(8)
        await sock.send(b'Message')
        data = await sock.recv(8192)
        await sock.close()
        await serv.cancel()

    async def main():
       serv = await spawn(tcp_server('',25000, handler))
       await spawn(test_client(('localhost', 25000), serv))

    kernel.run(main())

    assert results == [
            'handler start',
            ('client', b'Message'),
            'handler done'
            ]

def test_filestream_blocking(kernel):
    '''
    Test of exposing a socket in blocking mode
    '''
    results = []

    def sync_client(f):
        assert isinstance(f, io.RawIOBase)
        data = f.read(8192)
        results.append(('client', data))
        f.write(data)

    async def handler(client, addr):
        results.append('handler start')
        await client.send(b'OK')
        s = client.makefile('rwb', buffering=0)
        with s.blocking() as f:
            await run_in_thread(sync_client, f)
        results.append('handler done')

    async def test_client(address, serv):
        sock = socket(AF_INET, SOCK_STREAM)
        await sock.connect(address)
        await sock.recv(8)
        await sock.send(b'Message')
        data = await sock.recv(8192)
        await sock.close()
        await serv.cancel()

    async def main():
       serv = await spawn(tcp_server('',25000, handler))
       await spawn(test_client(('localhost', 25000), serv))

    kernel.run(main())

    assert results == [
            'handler start',
            ('client', b'Message'),
            'handler done'
            ]

def test_readall(kernel):
    results = []

    async def handler(client, addr):
        results.append('handler start')
        await client.send(b'OK')
        s = client.as_stream()
        data = await s.readall()
        results.append(data)
        results.append('handler done')

    async def test_client(address, serv):
        sock = socket(AF_INET, SOCK_STREAM)
        await sock.connect(address)
        await sock.recv(8)
        await sock.send(b'Msg1\n')
        await sleep(0.1)
        await sock.send(b'Msg2\n')
        await sleep(0.1)
        await sock.send(b'Msg3\n')
        await sleep(0.1)
        await sock.close()
        await serv.cancel()

    async def main():
       serv = await spawn(tcp_server('',25000, handler))
       await spawn(test_client(('localhost', 25000), serv))

    kernel.run(main())

    assert results == [
            'handler start',
            b'Msg1\nMsg2\nMsg3\n',
            'handler done'
            ]


def test_read_exactly(kernel):
    results = []

    async def handler(client, addr):
        results.append('handler start')
        await client.send(b'OK')
        s = client.as_stream()
        for n in range(3):
            results.append(await s.read_exactly(5))
        results.append('handler done')

    async def test_client(address, serv):
        sock = socket(AF_INET, SOCK_STREAM)
        await sock.connect(address)
        await sock.recv(8)
        await sock.send(b'Msg1\nMsg2\nMsg3\n')
        await sock.close()
        await serv.cancel()

    async def main():
       serv = await spawn(tcp_server('',25000, handler))
       await spawn(test_client(('localhost', 25000), serv))

    kernel.run(main())

    assert results == [
            'handler start',
            b'Msg1\n',
            b'Msg2\n',
            b'Msg3\n',
            'handler done'
            ]

def test_readline(kernel):
    results = []

    async def handler(client, addr):
        results.append('handler start')
        await client.send(b'OK')
        s = client.as_stream()
        for n in range(3):
            results.append(await s.readline())
        results.append('handler done')

    async def test_client(address, serv):
        sock = socket(AF_INET, SOCK_STREAM)
        await sock.connect(address)
        await sock.recv(8)
        await sock.send(b'Msg1\nMsg2\nMsg3\n')
        await sock.close()
        await serv.cancel()

    async def main():
       serv = await spawn(tcp_server('',25000, handler))
       await spawn(test_client(('localhost', 25000), serv))

    kernel.run(main())

    assert results == [
            'handler start',
            b'Msg1\n',
            b'Msg2\n',
            b'Msg3\n',
            'handler done'
            ]

def test_readlines(kernel):
    results = []

    async def handler(client, addr):
        await client.send(b'OK')
        results.append('handler start')
        s = client.as_stream()
        results.extend(await s.readlines())
        results.append('handler done')

    async def test_client(address, serv):
        sock = socket(AF_INET, SOCK_STREAM)
        await sock.connect(address)
        await sock.recv(8)
        await sock.send(b'Msg1\nMsg2\nMsg3\n')
        await sock.close()
        await serv.cancel()

    async def main():
       serv = await spawn(tcp_server('',25000, handler))
       await spawn(test_client(('localhost', 25000), serv))

    kernel.run(main())

    assert results == [
            'handler start',
            b'Msg1\n',
            b'Msg2\n',
            b'Msg3\n',
            'handler done'
            ]

def test_writelines(kernel):
    results = []

    async def handler(client, addr):
        results.append('handler start')
        await client.send(b'OK')
        s = client.as_stream()
        results.append(await s.readall())
        results.append('handler done')

    async def test_client(address, serv):
        sock = socket(AF_INET, SOCK_STREAM)
        await sock.connect(address)
        await sock.recv(8)
        s = sock.as_stream()
        await s.writelines([b'Msg1\n', b'Msg2\n', b'Msg3\n'])
        await sock.close()
        await serv.cancel()

    async def main():
       serv = await spawn(tcp_server('',25000, handler))
       await spawn(test_client(('localhost', 25000), serv))

    kernel.run(main())

    assert results == [
            'handler start',
            b'Msg1\nMsg2\nMsg3\n',
            'handler done'
            ]

def test_iterline(kernel):
    results = []

    async def handler(client, addr):
        results.append('handler start')
        await client.send(b'OK')
        s = client.as_stream()
        async for line in s:
            results.append(line)
        results.append('handler done')

    async def test_client(address, serv):
        sock = socket(AF_INET, SOCK_STREAM)
        await sock.connect(address)
        await sock.recv(8)
        await sock.send(b'Msg1\nMsg2\nMsg3\n')
        await sock.close()
        await serv.cancel()

    async def main():
       serv = await spawn(tcp_server('',25000, handler))
       await spawn(test_client(('localhost', 25000), serv))

    kernel.run(main())

    assert results == [
            'handler start',
            b'Msg1\n',
            b'Msg2\n',
            b'Msg3\n',
            'handler done'
            ]



# curio/subprocess.py
#
# A curio-compatible standin for the subprocess module.  Provides
# asynchronous compatible versions of Popen(), check_output(),
# and run() functions.

__all__ = [ 'run', 'Popen', 'CompletedProcess', 'CalledProcessError',
            'TimeoutExpired', 'SubprocessError', 'check_output',
            'PIPE', 'STDOUT', 'DEVNULL' ]

import subprocess

from subprocess import (
    CompletedProcess,
    SubprocessError, 
    CalledProcessError,
    TimeoutExpired,
    PIPE,
    STDOUT,
    DEVNULL,
    )

from .task import spawn, sleep, timeout_after
from .errors import TaskTimeout
from .io import FileStream

class Popen(object):
    '''
    Curio wrapper around the Popen class from the subprocess module. All of the
    methods from subprocess.Popen should be available, but the associated file
    objects for stdin, stdout, stderr have been replaced by async versions.
    Certain blocking operations (e.g., wait() and communicate()) have been
    replaced by async compatible implementations.
    '''
    def __init__(self, args, **kwargs):
        if 'universal_newlines' in kwargs:
            raise RuntimeError('universal_newlines argument not supported')

        self._popen = subprocess.Popen(args, **kwargs)
        if self._popen.stdin:
            self.stdin = FileStream(self._popen.stdin)
        if self._popen.stdout:
            self.stdout = FileStream(self._popen.stdout)
        if self._popen.stderr:
            self.stderr = FileStream(self._popen.stderr)

    def __getattr__(self, name):
        return getattr(self._popen, name)

    async def wait(self, timeout=None):
        async def waiter():
            while True:
                retcode = self._popen.poll()
                if retcode is not None:
                    return retcode
                await sleep(0.0005)
        task = await spawn(waiter())
        try:
            return await timeout_after(timeout, task.join())
        except TaskTimeout:
            await task.cancel()
            raise TimeoutExpired(self.args, timeout) from None

    async def communicate(self, input=b'', timeout=None):
        stdout_task = await spawn(self.stdout.readall()) if self.stdout else None
        stderr_task = await spawn(self.stderr.readall()) if self.stderr else None
        try:
            if input:
                await timeout_after(timeout, self.stdin.write(input))
                await self.stdin.close()
                self.stdin = None

            # Collect the output from the workers
            stdout = await timeout_after(timeout, stdout_task.join()) if stdout_task else b''
            stderr = await timeout_after(timeout, stderr_task.join()) if stderr_task else b''
            return (stdout, stderr)
        except TaskTimeout:
            await stdout_task.cancel()
            await stderr_task.cancel()
            raise

    def __enter__(self):
        raise RuntimeError('Use async with')

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        if self.stdout:
            await self.stdout.close()

        if self.stderr:
            await self.stderr.close()

        if self.stdin:
            await self.stdin.close()

        # Wait for the process to terminate
        await self.wait()

async def run(args, *, stdin=None, input=None, stdout=None, stderr=None, shell=False, timeout=None, check=False):
    '''
    Curio-compatible version of subprocess.run()
    '''
    if input:
        stdin = subprocess.PIPE
    else:
        stdin = None

    async with Popen(args, stdin=stdin, stdout=stdout, stderr=stderr, shell=shell) as process:
        try:
            stdout, stderr = await process.communicate(input, timeout)
        except TaskTimeout:
            process.kill()
            stdout, stderr = await process.communicate()
            raise TimeoutExpired(process.args, timeout, output=stdout, stderr=stderr)
        except:
            process.kill()
            raise

    retcode = process.poll()
    if check and retcode:
        raise CalledProcessError(retcode, process.args,
                                 output=stdout, stderr=stderr)
    return CompletedProcess(process.args, retcode, stdout, stderr)

async def check_output(args, *, stdin=None, stderr=None, shell=False, input=None, timeout=None):
     '''
     Curio compatible version of subprocess.check_output()
     '''
     out = await run(args, stdout=PIPE, stdin=stdin, stderr=stderr, shell=shell, 
                     timeout=timeout, check=True, input=input)
     return out.stdout



# Echo server with cancellation and signal handling

import signal
from curio import run, spawn, SignalSet, CancelledError, tcp_server, current_task

clients = set()

async def echo_client(client, addr):
    task = await current_task()
    clients.add(task)
    print('Connection from', addr)
    try:
        while True:
            data = await client.recv(1000)
            if not data:
                break
            await client.sendall(data)
        print('Connection closed')
    except CancelledError:
        await client.sendall(b'Server going down\n')
    finally:
        clients.remove(task)
    
async def main(host, port):
    while True:
        async with SignalSet(signal.SIGHUP) as sigset:
            print('Starting the server')
            serv_task = await spawn(tcp_server(host, port, echo_client))
            await sigset.wait()
            print('Server shutting down')
            await serv_task.cancel()

            for task in list(clients):
                await task.cancel()

if __name__ == '__main__':
    run(main('', 25000))



# echoserv.py
#
# Echo server using the run_server() function

from curio import run, tcp_server

async def echo_client(client, addr):
    print('Connection from', addr)
    while True:
        data = await client.recv(1000)
        if not data:
            break
        await client.sendall(data)
    print('Connection closed')

if __name__ == '__main__':
    run(tcp_server('', 25000, echo_client))



# thread.py
#
# Performance test of submitting work to a thread pool

import time
import curio
from concurrent.futures import ThreadPoolExecutor
import asyncio

COUNT = 25000

def curio_test():
    async def main():
        for n in range(COUNT):
            await curio.run_in_thread(time.sleep, 0)
    start = time.time()
    curio.run(main())
    end = time.time()
    print('Curio:', end-start)

def future_test():
    pool = ThreadPoolExecutor()
    def main():
        for n in range(COUNT):
            f = pool.submit(time.sleep, 0)
            r = f.result()
    start = time.time()
    main()
    end =  time.time()
    print('concurrent.futures:', end-start)

def asyncio_test():
    pool = ThreadPoolExecutor()
    async def main(loop):
        for n in range(COUNT):
            r = await loop.run_in_executor(pool, time.sleep, 0)

    loop = asyncio.get_event_loop()
    start = time.time()
    loop.run_until_complete(asyncio.ensure_future(main(loop)))
    end = time.time()
    print('asyncio:', end-start)

def uvloop_test():
    try:
        import uvloop
    except ImportError:
        return

    pool = ThreadPoolExecutor()
    async def main(loop):
        for n in range(COUNT):
            r = await loop.run_in_executor(pool, time.sleep, 0)

    loop = uvloop.new_event_loop()
    asyncio.set_event_loop(loop)
    start = time.time()
    loop.run_until_complete(asyncio.ensure_future(main(loop)))
    end = time.time()
    print('uvloop:', end-start)
    
if __name__ == '__main__':
    asyncio_test()
    uvloop_test()
    future_test()
    curio_test()




        



from gevent.server import StreamServer

# this handler will be run for each incoming connection in a dedicated greenlet
def echo(socket, address):
    print('New connection from %s:%s' % address)
    while True:
        data = socket.recv(100000)
        if not data:
            break
        socket.sendall(data)
    socket.close()

if __name__ == '__main__':
    server = StreamServer(('0.0.0.0', 25000), echo)
    server.serve_forever()



# prodcons.py
# 
# Example of a producer/consumer setup with queues

import curio

async def producer(queue):
    for n in range(10):
        await queue.put(n)
    await queue.join()
    print('Producer done')

async def consumer(queue):
    while True:
        item = await queue.get()
        print('Consumer got', item)
        await queue.task_done()

async def main():
    q = curio.Queue()
    prod_task = await curio.spawn(producer(q))
    cons_task = await curio.spawn(consumer(q))
    await prod_task.join()
    await cons_task.cancel()

if __name__ == '__main__':
    curio.run(main())



# ssl_http
#
# An example of a simple SSL server.  To test, connect via browser

import curio
from curio import ssl
import time

KEYFILE = "ssl_test_rsa"    # Private key
CERTFILE = "ssl_test.crt"   # Certificate (self-signed)

async def handler(client, addr):
    s = client.as_stream()
    async for line in s:
        line = line.strip()
        if not line:
            break
        print(line)

    await s.write(
b'''HTTP/1.0 200 OK\r
Content-type: text/plain\r
\r
If you're seeing this, it probably worked. Yay!
''')
    await s.write(time.asctime().encode('ascii'))
    await client.close()

if __name__ == '__main__':
    ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ssl_context.load_cert_chain(certfile=CERTFILE, keyfile=KEYFILE)
    print('Connect to https://localhost:10000 to see if it works')
    curio.run(curio.tcp_server('', 10000, handler, ssl=ssl_context))



# ssl_echo
#
# An example of a simple SSL echo client.  Use ssl_echo.py for the server.

import curio
from curio import ssl
from curio import network
from curio import socket

async def main(host, port):
    ssl_context = ssl.create_default_context()
    ssl_context.check_hostname = False
    ssl_context.verify_mode = ssl.CERT_NONE
    sock = await network.open_connection(host, port, ssl=True, server_hostname=None)
    for i in range(1000):
        msg = ('Message %d' % i).encode('ascii')
        print(msg)
        await sock.sendall(msg)
        resp = await sock.recv(1000)
        assert msg == resp
    await sock.close()

if __name__ == '__main__':
    curio.run(main('localhost', 10000))
